import pytest
from eth_utils import to_wei

from tests.utils import ZERO_ADDRESS

TOKEN_NAME = "Vypercoin"
TOKEN_SYMBOL = "FANG"
TOKEN_DECIMALS = 18
TOKEN_INITIAL_SUPPLY = 21 * 10**6
TOKEN_TOTAL_SUPPLY = TOKEN_INITIAL_SUPPLY * (10**TOKEN_DECIMALS)


@pytest.fixture(scope="module")
def erc20(get_contract):
    with open("examples/tokens/ERC20.vy") as f:
        contract_code = f.read()
    return get_contract(
        contract_code, *[TOKEN_NAME, TOKEN_SYMBOL, TOKEN_DECIMALS, TOKEN_INITIAL_SUPPLY]
    )


@pytest.fixture(scope="module")
def market_maker(get_contract, erc20):
    with open("examples/market_maker/on_chain_market_maker.vy") as f:
        contract_code = f.read()
    return get_contract(contract_code)


def test_initial_state(market_maker):
    assert market_maker.totalEthQty() == 0
    assert market_maker.totalTokenQty() == 0
    assert market_maker.invariant() == 0
    assert market_maker.owner() == ZERO_ADDRESS


def test_initiate(env, market_maker, erc20, tx_failed):
    a0 = env.accounts[0]
    ether, ethers = to_wei(1, "ether"), to_wei(2, "ether")
    env.set_balance(a0, ethers * 2)
    erc20.approve(market_maker.address, ethers)
    market_maker.initiate(erc20.address, ether, value=ethers)
    assert market_maker.totalEthQty() == ethers
    assert market_maker.totalTokenQty() == ether
    assert market_maker.invariant() == 2 * 10**36
    assert market_maker.owner() == a0
    assert erc20.name() == TOKEN_NAME
    assert erc20.decimals() == TOKEN_DECIMALS

    # Initiate cannot be called twice
    with tx_failed():
        market_maker.initiate(erc20.address, ether, value=ethers)


def test_eth_to_tokens(env, market_maker, erc20):
    a0, a1 = env.accounts[:2]
    env.set_balance(a0, to_wei(2, "ether"))
    erc20.approve(market_maker.address, to_wei(2, "ether"))
    market_maker.initiate(erc20.address, to_wei(1, "ether"), value=to_wei(2, "ether"))
    assert erc20.balanceOf(market_maker.address) == to_wei(1, "ether")
    assert erc20.balanceOf(a1) == 0
    assert market_maker.totalTokenQty() == to_wei(1, "ether")
    assert market_maker.totalEthQty() == to_wei(2, "ether")

    env.set_balance(a1, 100)
    market_maker.ethToTokens(value=100, sender=a1)
    assert erc20.balanceOf(market_maker.address) == 999999999999999950
    assert erc20.balanceOf(a1) == 50
    assert market_maker.totalTokenQty() == 999999999999999950
    assert market_maker.totalEthQty() == 2000000000000000100


def test_tokens_to_eth(env, market_maker, erc20):
    a1 = env.accounts[1]
    a1_balance_before = to_wei(2, "ether")
    env.set_balance(a1, a1_balance_before)

    erc20.transfer(a1, to_wei(2, "ether"))
    erc20.approve(market_maker.address, to_wei(2, "ether"), sender=a1)
    market_maker.initiate(erc20.address, to_wei(1, "ether"), value=to_wei(2, "ether"), sender=a1)
    assert env.get_balance(market_maker.address) == to_wei(2, "ether")
    # sent 2 eth, with initiate.
    assert env.get_balance(a1) == a1_balance_before - to_wei(2, "ether")
    assert market_maker.totalTokenQty() == to_wei(1, "ether")

    erc20.approve(market_maker.address, to_wei(1, "ether"), sender=a1)
    market_maker.tokensToEth(to_wei(1, "ether"), sender=a1)
    # 1 eth less in market.
    assert env.get_balance(market_maker.address) == to_wei(1, "ether")
    # got 1 eth back, for trade.
    assert env.get_balance(a1) == a1_balance_before - to_wei(1, "ether")
    # Tokens increased by 1
    assert market_maker.totalTokenQty() == to_wei(2, "ether")
    assert market_maker.totalEthQty() == to_wei(1, "ether")


def test_owner_withdraw(env, market_maker, erc20, tx_failed):
    a0, a1 = env.accounts[:2]
    a0_balance_before = to_wei(10, "ether")
    env.set_balance(a0, a0_balance_before)
    # Approve 2 eth transfers.
    erc20.approve(market_maker.address, to_wei(2, "ether"))
    # Initiate market with 2 eth value.
    market_maker.initiate(erc20.address, to_wei(1, "ether"), value=to_wei(2, "ether"))
    # 2 eth was sent to market_maker contract.
    assert env.get_balance(a0) == a0_balance_before - to_wei(2, "ether")
    # a0's balance is locked up in market_maker contract.
    assert erc20.balanceOf(a0) == TOKEN_TOTAL_SUPPLY - to_wei(1, "ether")

    # Only owner can call ownerWithdraw
    with tx_failed():
        market_maker.ownerWithdraw(sender=a1)
    market_maker.ownerWithdraw()
    assert env.get_balance(a0) == a0_balance_before  # Eth balance restored.
    assert erc20.balanceOf(a0) == TOKEN_TOTAL_SUPPLY  # Tokens returned to a0.



import pytest

from vyper.compiler import compile_code
from vyper.utils import method_id


def test_simple_export(make_input_bundle, get_contract):
    lib1 = """
@external
def foo() -> uint256:
    return 5
    """
    main = """
import lib1

exports: lib1.foo
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    c = get_contract(main, input_bundle=input_bundle)

    assert c.foo() == 5


def test_export_with_state(make_input_bundle, get_contract):
    lib1 = """
counter: uint256

@external
def foo() -> uint256:
    return self.counter
    """
    main = """
import lib1

initializes: lib1
exports: lib1.foo

@deploy
def __init__():
    lib1.counter = 99
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    c = get_contract(main, input_bundle=input_bundle)

    assert c.foo() == 99


def test_variable_decl_exports(make_input_bundle, get_contract):
    lib1 = """
counter: public(uint256)
FOO: public(immutable(uint256))
BAR: public(constant(uint256)) = 3

@deploy
def __init__():
    self.counter = 1
    FOO = 2
    """
    main = """
import lib1

initializes: lib1
exports: (
    lib1.counter,
    lib1.FOO,
    lib1.BAR,
)

@deploy
def __init__():
    lib1.__init__()
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    c = get_contract(main, input_bundle=input_bundle)

    assert c.counter() == 1
    assert c.FOO() == 2
    assert c.BAR() == 3


def test_not_exported(make_input_bundle, get_contract):
    # test that non-exported functions are not in the selector table
    lib1 = """
@external
def foo() -> uint256:
    return 100

@external
def bar() -> uint256:
    return 101
    """
    main = """
import lib1

exports: lib1.foo

@external
def __default__() -> uint256:
    return 127
    """
    caller_code = """
interface Foo:
    def foo() -> uint256: nonpayable
    def bar() -> uint256: nonpayable

@external
def call_bar(foo: Foo) -> uint256:
    return extcall foo.bar()
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    c = get_contract(main, input_bundle=input_bundle)
    caller = get_contract(caller_code)

    assert caller.call_bar(c.address) == 127  # default return value


def test_nested_export(make_input_bundle, get_contract):
    lib1 = """
@external
def foo() -> uint256:
    return 5
    """
    lib2 = """
import lib1
    """
    main = """
import lib2

exports: lib2.lib1.foo
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})
    c = get_contract(main, input_bundle=input_bundle)

    assert c.foo() == 5


def test_transitive_export(make_input_bundle, get_contract):
    lib1 = """
@external
def foo() -> uint256:
    return 5
    """
    lib2 = """
import lib1
exports: lib1.foo
    """
    main = """
import lib2

exports: lib2.foo
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})
    c = get_contract(main, input_bundle=input_bundle)

    assert c.foo() == 5


@pytest.fixture
def simple_library(make_input_bundle):
    ifoo = """
@external
def foo() -> uint256:
    ...

@external
def bar() -> uint256:
    ...
    """
    ibar = """
@external
def bar() -> uint256:
    ...

@external
def qux() -> uint256:
    ...
    """
    lib1 = """
import ifoo
import ibar

implements: ifoo
implements: ibar

@external
def foo() -> uint256:
    return 1

@external
def bar() -> uint256:
    return 2

@external
def qux() -> uint256:
    return 3
    """
    return make_input_bundle({"lib1.vy": lib1, "ifoo.vyi": ifoo, "ibar.vyi": ibar})


@pytest.fixture
def send_failing_tx_to_signature(env, tx_failed):
    def _send_transaction(c, method_sig):
        data = method_id(method_sig)
        with tx_failed():
            env.message_call(c.address, data=data)

    return _send_transaction


def test_exports_interface_simple(get_contract, simple_library):
    main = """
import lib1

exports: lib1.__interface__
    """
    c = get_contract(main, input_bundle=simple_library)
    assert c.foo() == 1
    assert c.bar() == 2
    assert c.qux() == 3


def test_exports_interface2(get_contract, send_failing_tx_to_signature, simple_library):
    main = """
import lib1

exports: lib1.ifoo
    """
    out = compile_code(
        main, output_formats=["abi"], contract_path="main.vy", input_bundle=simple_library
    )
    fnames = [item["name"] for item in out["abi"]]
    assert fnames == ["foo", "bar"]

    c = get_contract(main, input_bundle=simple_library)
    assert c.foo() == 1
    assert c.bar() == 2
    assert not hasattr(c, "qux")
    send_failing_tx_to_signature(c, "qux()")


def test_exported_fun_part_of_interface(get_contract, make_input_bundle):
    main = """
import lib2

exports: lib2.__interface__
    """
    lib1 = """
@external
def bar() -> uint256:
    return 1
    """
    lib2 = """
import lib1

@external
def foo() -> uint256:
    return 2

exports: lib1.bar
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})
    c = get_contract(main, input_bundle=input_bundle)
    assert c.bar() == 1
    assert c.foo() == 2


def test_imported_module_not_part_of_interface(
    send_failing_tx_to_signature, get_contract, make_input_bundle
):
    main = """
import lib2

exports: lib2.__interface__
    """
    lib1 = """
@external
def bar() -> uint256:
    return 1
    """
    lib2 = """
import lib1

@external
def foo() -> uint256:
    return 2
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})
    c = get_contract(main, input_bundle=input_bundle)
    assert c.foo() == 2
    send_failing_tx_to_signature(c, "bar()")


def test_export_unimplemented_function(
    send_failing_tx_to_signature, get_contract, make_input_bundle
):
    ifoo = """
@external
def foo() -> uint256:
    ...
    """
    lib1 = """
import ifoo
implements: ifoo

@external
def foo() -> uint256:
    return 1

@external
def bar() -> uint256:
    return 2
    """
    main = """
import lib1

exports: lib1.ifoo
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "ifoo.vyi": ifoo})
    c = get_contract(main, input_bundle=input_bundle)
    assert c.foo() == 1
    send_failing_tx_to_signature(c, "bar()")


# sanity check that when multiple modules implement an interface, the
# correct one (specified by the user) gets selected for export.
def test_export_interface_multiple_choices(get_contract, make_input_bundle):
    ifoo = """
@external
def foo() -> uint256:
    ...
    """
    lib1 = """
import ifoo
implements: ifoo

@external
def foo() -> uint256:
    return 1
    """
    lib2 = """
import ifoo
implements: ifoo

@external
def foo() -> uint256:
    return 2
    """
    main = """
import lib1
import lib2

exports: lib1.ifoo
    """
    main2 = """
import lib1
import lib2

exports: lib2.ifoo
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2, "ifoo.vyi": ifoo})

    c = get_contract(main, input_bundle=input_bundle)
    assert c.foo() == 1

    c = get_contract(main2, input_bundle=input_bundle)
    assert c.foo() == 2


def test_export_module_with_init(get_contract, make_input_bundle):
    lib1 = """
@deploy
def __init__():
    pass

@external
def foo() -> uint256:
    return 1
    """
    main = """
import lib1

exports: lib1.__interface__
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    c = get_contract(main, input_bundle=input_bundle)
    assert c.foo() == 1


def test_export_module_with_getter(get_contract, make_input_bundle):
    lib1 = """
counter: public(uint256)

@external
def foo():
    self.counter += 1
    """
    main = """
import lib1

initializes: lib1
exports: lib1.__interface__

@deploy
def __init__():
    lib1.counter = 100
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    c = get_contract(main, input_bundle=input_bundle)
    assert c.counter() == 100
    c.foo()
    assert c.counter() == 101


def test_export_module_with_default(env, get_contract, make_input_bundle):
    lib1 = """
counter: public(uint256)

@external
def foo() -> uint256:
    return 1

@external
def __default__():
    self.counter += 1
    """
    main = """
import lib1
initializes: lib1

@deploy
def __init__():
    lib1.counter = 5

exports: lib1.__interface__
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    c = get_contract(main, input_bundle=input_bundle)
    assert c.foo() == 1
    assert c.counter() == 5
    # call `c.__default__()`
    env.message_call(c.address)
    assert c.counter() == 6



from vyper.codegen.core import (
    STORE,
    add_ofst,
    get_dyn_array_count,
    get_element_ptr,
    is_tuple_like,
    make_setter,
    zero_pad,
)
from vyper.codegen.ir_node import IRnode
from vyper.evm.address_space import MEMORY
from vyper.exceptions import CompilerPanic
from vyper.semantics.types import DArrayT, SArrayT, _BytestringT
from vyper.semantics.types.shortcuts import UINT256_T


def _is_complex_type(typ):
    # Tuples, Structs, and SArrays follow the same code path
    return is_tuple_like(typ) or isinstance(typ, (SArrayT))


# turn an ir node into a list, based on its type.
def _deconstruct_complex_type(ir_node):
    ir_t = ir_node.typ
    assert _is_complex_type(ir_t)

    if is_tuple_like(ir_t):
        ks = ir_t.tuple_keys()
    else:
        ks = [IRnode.from_list(i, UINT256_T) for i in range(ir_t.count)]

    ret = []
    for k in ks:
        ret.append(get_element_ptr(ir_node, k, array_bounds_check=False))
    return ret


# encode a child element of a complex type
def _encode_child_helper(buf, child, static_ofst, dyn_ofst, context):
    child_abi_t = child.typ.abi_type

    static_loc = add_ofst(IRnode.from_list(buf), static_ofst)

    ret = ["seq"]

    if not child_abi_t.is_dynamic():
        # easy
        _bufsz = child_abi_t.size_bound()
        ret.append(abi_encode(static_loc, child, context, _bufsz, returns_len=False))
    else:
        # hard
        ret.append(["mstore", static_loc, dyn_ofst])

        # TODO optimize: special case where there is only one dynamic
        # member, the location is statically known.
        child_dst = ["add", buf, dyn_ofst]

        _bufsz = child_abi_t.size_bound()
        child_len = abi_encode(child_dst, child, context, _bufsz, returns_len=True)

        # increment dyn ofst for return_len
        # (optimization note:
        #   if non-returning and this is the last dyn member in
        #   the tuple, this set can be elided.)
        ret.append(["set", dyn_ofst, ["add", dyn_ofst, child_len]])

    return ret


def _encode_dyn_array_helper(dst, ir_node, context):
    # if it's a literal, first serialize to memory as we
    # don't have a compile-time abi encoder
    # TODO handle this upstream somewhere
    if ir_node.value == "multi":
        buf = context.new_internal_variable(dst.typ)
        buf = IRnode.from_list(buf, typ=dst.typ, location=MEMORY)
        _bufsz = dst.typ.abi_type.size_bound()
        return [
            "seq",
            make_setter(buf, ir_node),
            ["set", "dyn_ofst", abi_encode(dst, buf, context, _bufsz, returns_len=True)],
        ]

    subtyp = ir_node.typ.value_type
    child_abi_t = subtyp.abi_type

    ret = ["seq"]

    len_ = get_dyn_array_count(ir_node)
    with len_.cache_when_complex("len") as (b, len_):
        # set the length word
        ret.append(STORE(dst, len_))

        # prepare the loop
        t = UINT256_T
        i = IRnode.from_list(context.fresh_varname("ix"), typ=t)

        # offset of the i'th element in ir_node
        child_location = get_element_ptr(ir_node, i, array_bounds_check=False)

        # offset of the i'th element in dst
        dst = add_ofst(dst, 32)  # jump past length word
        static_elem_size = child_abi_t.embedded_static_size()
        static_ofst = ["mul", i, static_elem_size]
        loop_body = _encode_child_helper(
            dst, child_location, static_ofst, "dyn_child_ofst", context
        )
        loop = ["repeat", i, 0, len_, ir_node.typ.count, loop_body]

        x = ["seq", loop, "dyn_child_ofst"]
        start_dyn_ofst = ["mul", len_, static_elem_size]
        run_children = ["with", "dyn_child_ofst", start_dyn_ofst, x]
        new_dyn_ofst = ["add", "dyn_ofst", run_children]
        # size of dynarray is size of encoded children + size of the length word
        # TODO optimize by adding 32 to the initial value of dyn_ofst
        new_dyn_ofst = ["add", 32, new_dyn_ofst]
        ret.append(["set", "dyn_ofst", new_dyn_ofst])

        return b.resolve(ret)


def abi_encoding_matches_vyper(typ):
    """
    returns True if the ABI encoding matches vyper's memory encoding of
    a type, otherwise False
    """
    return not typ.abi_type.is_dynamic()


# assume dst is a pointer to a buffer located in memory which has at
# least static_size + dynamic_size_bound allocated.
# The basic strategy is this:
#   First, it is helpful to keep track of what variables are location
#   dependent and which are location independent (offsets). Independent
#   locations will be denoted with variables named `_ofst`.
#   We keep at least one stack variable to keep track of our location
#   in the dynamic section. We keep a compiler variable `static_ofst`
#   keeping track of our current offset from the beginning of the static
#   section. (And if the destination is not known at compile time, we
#   allocate a stack item `dst` to keep track of it). So, 1-2 stack
#   variables for each level of "nesting" (as defined in the spec).
#   For each element `elem` of the ir_node:
#   - If `elem` is static, write its value to `dst + static_ofst` and
#     increment `static_ofst` by the size of `elem`.
#   - If it is dynamic, ensure we have initialized a pointer (a stack
#     variable named `dyn_ofst` set to the start of the dynamic section
#     (i.e. static_size of ir_node). Write the 'tail' of `elem` to the
#     dynamic section, then write current `dyn_ofst` to `dst_loc`, and
#     then increment `dyn_ofst` by the number of bytes written. Note
#     that in this step we may recurse, and the child call should return
#     a stack item representing how many bytes were written.
#     WARNING: abi_encode(bytes) != abi_encode((bytes,)) (a tuple
#     with a single bytes member). The former is encoded as <len> <data>,
#     the latter is encoded as <ofst> <len> <data>.
# performance note: takes O(n^2) compilation time
# where n is depth of data type, could be optimized but unlikely
# that users will provide deeply nested data.
# returns_len is a calling convention parameter; if set to true,
# the abi_encode routine will push the output len onto the stack,
# otherwise it will return 0 items to the stack.
def abi_encode(dst, ir_node, context, bufsz, returns_len=False):
    # cast dst to the type of the input so that make_setter works
    dst = IRnode.from_list(dst, typ=ir_node.typ, location=MEMORY)

    abi_t = dst.typ.abi_type
    size_bound = abi_t.size_bound()

    assert isinstance(bufsz, int)
    if bufsz < size_bound:  # pragma: nocover
        raise CompilerPanic("buffer provided to abi_encode not large enough")

    if size_bound < dst.typ.memory_bytes_required:  # pragma: nocover
        raise CompilerPanic("Bad ABI size calc")

    annotation = f"abi_encode {ir_node.typ}"
    ir_ret = ["seq"]

    # fastpath: if there is no dynamic data, we can optimize the
    # encoding by using make_setter, since our memory encoding happens
    # to be identical to the ABI encoding.
    if abi_encoding_matches_vyper(ir_node.typ):
        # NOTE: make_setter handles changes of location and encoding
        ir_ret.append(make_setter(dst, ir_node))
        if returns_len:
            assert abi_t.embedded_static_size() == ir_node.typ.memory_bytes_required
            ir_ret.append(abi_t.embedded_static_size())
        return IRnode.from_list(ir_ret, annotation=annotation)

    # contains some computation, we need to only do it once.
    with ir_node.cache_when_complex("to_encode") as (b1, ir_node), dst.cache_when_complex(
        "dst"
    ) as (b2, dst):
        dyn_ofst = "dyn_ofst"  # current offset in the dynamic section

        if ir_node.typ._is_prim_word:
            ir_ret.append(make_setter(dst, ir_node))
        elif isinstance(ir_node.typ, _BytestringT):
            # TODO optimize out repeated ceil32 calculation
            ir_ret.append(make_setter(dst, ir_node))
            ir_ret.append(zero_pad(dst))
        elif isinstance(ir_node.typ, DArrayT):
            ir_ret.append(_encode_dyn_array_helper(dst, ir_node, context))
        elif _is_complex_type(ir_node.typ):
            static_ofst = 0
            elems = _deconstruct_complex_type(ir_node)
            for e in elems:
                encode_ir = _encode_child_helper(dst, e, static_ofst, dyn_ofst, context)
                ir_ret.extend(encode_ir)
                static_ofst += e.typ.abi_type.embedded_static_size()

        else:  # pragma: nocover
            raise CompilerPanic(f"unencodable type: {ir_node.typ}")

        # declare IR variables.
        if returns_len:
            if not abi_t.is_dynamic():
                ir_ret.append(abi_t.embedded_static_size())
            elif isinstance(ir_node.typ, _BytestringT):
                # for abi purposes, return zero-padded length
                calc_len = ["ceil32", ["add", 32, ["mload", dst]]]
                ir_ret.append(calc_len)
            elif abi_t.is_complex_type():
                ir_ret.append("dyn_ofst")
            else:  # pragma: nocover
                raise CompilerPanic(f"unknown type {ir_node.typ}")

        if abi_t.is_dynamic() and abi_t.is_complex_type():
            dyn_section_start = abi_t.static_size()
            ir_ret = ["with", dyn_ofst, dyn_section_start, ir_ret]
        else:
            pass  # skip dyn_ofst allocation if we don't need it

        return b1.resolve(b2.resolve(IRnode.from_list(ir_ret, annotation=annotation)))



from tests.evm_backends.base_env import EvmError


def test_unreachable_refund(env, get_contract, tx_failed):
    code = """
@external
def foo():
    assert msg.sender != msg.sender, UNREACHABLE
    """
    env.set_balance(env.deployer, 10**20)

    c = get_contract(code)
    gas_sent = 10**6
    with tx_failed(EvmError, exc_text=env.invalid_opcode_error):
        c.foo(gas=gas_sent, gas_price=10)

    assert env.last_result.gas_used == gas_sent  # Drains all gas sent per INVALID opcode


def test_basic_unreachable(env, get_contract, tx_failed):
    code = """
@external
def foo(val: int128) -> bool:
    assert val > 0, UNREACHABLE
    assert val == 2, UNREACHABLE
    return True
    """

    c = get_contract(code)

    assert c.foo(2) is True

    with tx_failed(EvmError, exc_text=env.invalid_opcode_error):
        c.foo(1)
    with tx_failed(EvmError, exc_text=env.invalid_opcode_error):
        c.foo(-1)
    with tx_failed(EvmError, exc_text=env.invalid_opcode_error):
        c.foo(-2)


def test_basic_call_unreachable(env, get_contract, tx_failed):
    code = """

@view
@internal
def _test_me(val: int128) -> bool:
    return val == 33

@external
def foo(val: int128) -> int128:
    assert self._test_me(val), UNREACHABLE
    return -123
    """

    c = get_contract(code)

    assert c.foo(33) == -123

    with tx_failed(EvmError, exc_text=env.invalid_opcode_error):
        c.foo(1)
    with tx_failed(EvmError, exc_text=env.invalid_opcode_error):
        c.foo(-1)


def test_raise_unreachable(env, get_contract, tx_failed):
    code = """
@external
def foo():
    raise UNREACHABLE
    """

    c = get_contract(code)

    with tx_failed(EvmError, exc_text=env.invalid_opcode_error):
        c.foo()



import pytest

from vyper.exceptions import StaticAssertionException
from vyper.utils import SizeLimits


def test_basic_repeater(get_contract):
    basic_repeater = """
@external
def repeat(z: int128) -> int128:
    x: int128 = 0
    for i: int128 in range(6):
        x = x + z
    return(x)
    """
    c = get_contract(basic_repeater)
    assert c.repeat(9) == 54


def test_range_bound(get_contract, tx_failed):
    code = """
@external
def repeat(n: uint256) -> uint256:
    x: uint256 = 0
    for i: uint256 in range(n, bound=6):
        x += i + 1
    return x
    """
    c = get_contract(code)
    for n in range(7):
        assert c.repeat(n) == sum(i + 1 for i in range(n))

    # check codegen inserts assertion for n greater than bound
    with tx_failed():
        c.repeat(7)


def test_range_bound_constant_end(get_contract, tx_failed):
    code = """
@external
def repeat(n: uint256) -> uint256:
    x: uint256 = 0
    for i: uint256 in range(n, 7, bound=6):
        x += i + 1
    return x
    """
    c = get_contract(code)
    for n in range(1, 5):
        assert c.repeat(n) == sum(i + 1 for i in range(n, 7))

    # check assertion for `start <= end`
    with tx_failed():
        c.repeat(8)
    # check assertion for `start + bound <= end`
    with tx_failed():
        c.repeat(0)


def test_range_bound_two_args(get_contract, tx_failed):
    code = """
@external
def repeat(n: uint256) -> uint256:
    x: uint256 = 0
    for i: uint256 in range(1, n, bound=6):
        x += i + 1
    return x
    """
    c = get_contract(code)
    for n in range(1, 8):
        assert c.repeat(n) == sum(i + 1 for i in range(1, n))

    # check assertion for `start <= end`
    with tx_failed():
        c.repeat(0)

    # check codegen inserts assertion for `start + bound <= end`
    with tx_failed():
        c.repeat(8)


def test_range_bound_two_runtime_args(get_contract, tx_failed):
    code = """
@external
def repeat(start: uint256, end: uint256) -> uint256:
    x: uint256 = 0
    for i: uint256 in range(start, end, bound=6):
        x += i
    return x
    """
    c = get_contract(code)
    for n in range(0, 7):
        assert c.repeat(0, n) == sum(range(0, n))
        assert c.repeat(n, n * 2) == sum(range(n, n * 2))

    # check assertion for `start <= end`
    with tx_failed():
        c.repeat(1, 0)
    with tx_failed():
        c.repeat(7, 0)
    with tx_failed():
        c.repeat(8, 7)

    # check codegen inserts assertion for `start + bound <= end`
    with tx_failed():
        c.repeat(0, 7)
    with tx_failed():
        c.repeat(14, 21)


def test_range_overflow(get_contract, tx_failed):
    code = """
@external
def get_last(start: uint256, end: uint256) -> uint256:
    x: uint256 = 0
    for i: uint256 in range(start, end, bound=6):
        x = i
    return x
    """
    c = get_contract(code)
    UINT_MAX = 2**256 - 1
    assert c.get_last(UINT_MAX, UINT_MAX) == 0  # initial value of x

    for n in range(1, 6):
        assert c.get_last(UINT_MAX - n, UINT_MAX) == UINT_MAX - 1

    # check for `start + bound <= end`, overflow cases
    for n in range(1, 7):
        with tx_failed():
            c.get_last(UINT_MAX - n, 0)
        with tx_failed():
            c.get_last(UINT_MAX, UINT_MAX - n)


def test_digit_reverser(get_contract):
    digit_reverser = """
@external
def reverse_digits(x: int128) -> int128:
    dig: int128[6] = [0, 0, 0, 0, 0, 0]
    z: int128 = x
    for i: uint256 in range(6):
        dig[i] = z % 10
        z = z // 10
    o: int128 = 0
    for i: uint256 in range(6):
        o = o * 10 + dig[i]
    return o

    """

    c = get_contract(digit_reverser)
    assert c.reverse_digits(123456) == 654321


def test_more_complex_repeater(get_contract):
    more_complex_repeater = """
@external
def repeat() -> int128:
    out: int128 = 0
    for i: uint256 in range(6):
        out = out * 10
        for j: int128 in range(4):
            out = out + j
    return(out)
    """

    c = get_contract(more_complex_repeater)
    assert c.repeat() == 666666


@pytest.mark.parametrize("typ", ["int128", "uint256"])
def test_offset_repeater(get_contract, typ):
    offset_repeater = f"""
@external
def sum() -> {typ}:
    out: {typ} = 0
    for i: {typ} in range(80, 121):
        out = out + i
    return out
    """

    c = get_contract(offset_repeater)
    assert c.sum() == 4100


@pytest.mark.parametrize("typ", ["int128", "uint256"])
def test_offset_repeater_2(get_contract, typ):
    offset_repeater_2 = f"""
@external
def sum(frm: {typ}, to: {typ}) -> {typ}:
    out: {typ} = 0
    for i: {typ} in range(frm, frm + 101, bound=101):
        if i == to:
            break
        out = out + i
    return out
    """

    c = get_contract(offset_repeater_2)
    assert c.sum(100, 99999) == 15150
    assert c.sum(70, 131) == 6100


def test_loop_call_priv(get_contract):
    code = """
@internal
def _bar() -> bool:
    return True

@external
def foo() -> bool:
    for i: uint256 in range(3):
        self._bar()
    return True
    """

    c = get_contract(code)
    assert c.foo() is True


@pytest.mark.parametrize("typ", ["int128", "uint256"])
def test_return_inside_repeater(get_contract, typ):
    code = f"""
@internal
def _final(a: {typ}) -> {typ}:
    for i: {typ} in range(10):
        for j: {typ} in range(10):
            if j > 5:
                if i > a:
                    return i
    return 31337

@internal
def _middle(a: {typ}) -> {typ}:
    b: {typ} = self._final(a)
    return b

@external
def foo(a: {typ}) -> {typ}:
    b: {typ} = self._middle(a)
    return b
    """

    c = get_contract(code)
    assert c.foo(6) == 7
    assert c.foo(100) == 31337


@pytest.mark.parametrize("typ", ["uint8", "int128", "uint256"])
def test_for_range_edge(get_contract, typ):
    """
    Check that we can get to the upper range of an integer.
    Note that to avoid overflow in the bounds check for range(),
    we need to calculate i+1 inside the loop.
    """
    code = f"""
@external
def test():
    found: bool = False
    x: {typ} = max_value({typ})
    for i: {typ} in range(x - 1, x, bound=1):
        if i + 1 == max_value({typ}):
            found = True
    assert found

    found = False
    x = max_value({typ}) - 1
    for i: {typ} in range(x - 1, x + 1, bound=2):
        if i + 1 == max_value({typ}):
            found = True
    assert found
    """
    c = get_contract(code)
    c.test()


@pytest.mark.parametrize("typ", ["uint8", "int128", "uint256"])
def test_for_range_oob_compile_time_check(get_contract, tx_failed, typ, experimental_codegen):
    code = f"""
@external
def test():
    x: {typ} = max_value({typ})
    for i: {typ} in range(x, x + 2, bound=2):
        pass
    """
    if not experimental_codegen:
        return
    with pytest.raises(StaticAssertionException):
        get_contract(code)


@pytest.mark.parametrize(
    "typ, max_value",
    [
        ("uint8", SizeLimits.MAX_UINT8),
        ("int128", SizeLimits.MAX_INT128),
        ("uint256", SizeLimits.MAX_UINT256),
    ],
)
def test_for_range_oob_runtime_check(get_contract, tx_failed, typ, max_value):
    code = f"""
@external
def test(x: {typ}):
    for i: {typ} in range(x, x + 2, bound=2):
        pass
    """
    c = get_contract(code)
    with tx_failed():
        c.test(max_value)


@pytest.mark.parametrize("typ", ["int128", "uint256"])
def test_return_inside_nested_repeater(get_contract, typ):
    code = f"""
@internal
def _final(a: {typ}) -> {typ}:
    for i: {typ} in range(10):
        for x: {typ} in range(10):
            if i + x > a:
                return i + x
    return 31337

@internal
def _middle(a: {typ}) -> {typ}:
    b: {typ} = self._final(a)
    return b

@external
def foo(a: {typ}) -> {typ}:
    b: {typ} = self._middle(a)
    return b
    """

    c = get_contract(code)
    assert c.foo(14) == 15
    assert c.foo(100) == 31337


@pytest.mark.parametrize("typ", ["int128", "uint256"])
@pytest.mark.parametrize("val", range(20))
def test_return_void_nested_repeater(get_contract, typ, val):
    code = f"""
result: {typ}
@internal
def _final(a: {typ}):
    for i: {typ} in range(10):
        for x: {typ} in range(10):
            if i + x > a:
                self.result = i + x
                return
    self.result = 31337

@internal
def _middle(a: {typ}):
    self._final(a)

@external
def foo(a: {typ}) -> {typ}:
    self._middle(a)
    return self.result
    """
    c = get_contract(code)
    if val + 1 >= 19:
        assert c.foo(val) == 31337
    else:
        assert c.foo(val) == val + 1


@pytest.mark.parametrize("typ", ["int128", "uint256"])
@pytest.mark.parametrize("val", range(20))
def test_external_nested_repeater(get_contract, typ, val):
    code = f"""
@external
def foo(a: {typ}) -> {typ}:
    for i: {typ} in range(10):
        for x: {typ} in range(10):
            if i + x > a:
                return i + x
    return 31337
    """
    c = get_contract(code)
    if val + 1 >= 19:
        assert c.foo(val) == 31337
    else:
        assert c.foo(val) == val + 1


@pytest.mark.parametrize("typ", ["int128", "uint256"])
@pytest.mark.parametrize("val", range(20))
def test_external_void_nested_repeater(get_contract, typ, val):
    # test return out of loop in void external function
    code = f"""
result: public({typ})
@external
def foo(a: {typ}):
    for i: {typ} in range(10):
        for x: {typ} in range(10):
            if i + x > a:
                self.result = i + x
                return
    self.result = 31337
    """
    c = get_contract(code)
    c.foo(val)
    if val + 1 >= 19:
        assert c.result() == 31337
    else:
        assert c.result() == val + 1


@pytest.mark.parametrize("typ", ["int128", "uint256"])
def test_breaks_and_returns_inside_nested_repeater(get_contract, typ):
    code = f"""
@internal
def _final(a: {typ}) -> {typ}:
    for i: {typ} in range(10):
        for x: {typ} in range(10):
            if a < 2:
                break
            return 6
        if a == 1:
            break
        return 31337

    return 666

@internal
def _middle(a: {typ}) -> {typ}:
    b: {typ} = self._final(a)
    return b

@external
def foo(a: {typ}) -> {typ}:
    b: {typ} = self._middle(a)
    return b
    """

    c = get_contract(code)
    assert c.foo(100) == 6
    assert c.foo(1) == 666
    assert c.foo(0) == 31337


def test_for_range_signed_int_overflow_runtime_check(get_contract, tx_failed, experimental_codegen):
    code = """
@external
def foo(_min:int256, _max: int256) -> DynArray[int256, 10]:
    res: DynArray[int256, 10] = empty(DynArray[int256, 10])
    x:int256 = _max
    y:int256 = _min+2
    for i:int256 in range(x,y , bound=10):
        res.append(i)
    return res
    """
    c = get_contract(code)
    with tx_failed():
        c.foo(SizeLimits.MIN_INT256, SizeLimits.MAX_INT256)


def test_for_range_signed_int_overflow_compile_time_check(
    get_contract, tx_failed, experimental_codegen
):
    code = """
@external
def foo() -> DynArray[int256, 10]:
    res: DynArray[int256, 10] = empty(DynArray[int256, 10])
    x:int256 = max_value(int256)
    y:int256 = min_value(int256)+2
    for i:int256 in range(x,y , bound=10):
        res.append(i)
    return res
    """
    if not experimental_codegen:
        return
    with pytest.raises(StaticAssertionException):
        get_contract(code)



import pytest

from vyper import compile_code
from vyper.exceptions import OverflowException, TypeMismatch

fail_list = [
    (
        """
@external
def foo():
    y: int128 = min(7, 0x1234567890123456789012345678901234567890)
    """,
        TypeMismatch,
    ),
    (
        """
@external
def foo():
    a: int16 = min(min_value(int16), max_value(int8))
    """,
        TypeMismatch,
    ),
    (
        """
@external
def foo():
   a: decimal = min(1.0, 18707220957835557353007165858768422651595.9365500928)
    """,
        OverflowException,
    ),
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_block_fail(bad_code, exc):
    with pytest.raises(exc):
        compile_code(bad_code)


valid_list = [
    """
FOO: constant(uint256) = 123
BAR: constant(uint256) = 456
BAZ: constant(uint256) = min(FOO, BAR)

@external
def foo():
    a: uint256 = BAZ
    """,
    """
FOO: constant(uint256) = 123
BAR: constant(uint256) = 456
BAZ: constant(uint256) = max(FOO, BAR)

@external
def foo():
    a: uint256 = BAZ
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_block_success(good_code):
    assert compile_code(good_code) is not None



import pytest

from vyper import compiler
from vyper.exceptions import UndeclaredDefinition

fail_list = [
    """
@external
def test1(b: uint256) -> uint256:
    a: uint256 = a + b
    return a
    """,
    """
@external
def test2(b: uint256, c: uint256) -> uint256:
    a: uint256 = a + b + c
    return a
    """,
    """
@external
def test3(b: int128, c: int128) -> int128:
    a: int128 = - a
    return a
    """,
    """
@external
def test4(b: bool) -> bool:
    a: bool = b or a
    return a
    """,
    """
@external
def test5(b: bool) -> bool:
    a: bool = a != b
    return a
    """,
    """
@external
def test6(b:bool, c: bool) -> bool:
    a: bool = (a and b) and c
    return a
    """,
    """
@external
def foo():
    throe = 2
    """,
    """
@external
def foo():
    x: int128 = bar(55)
    """,
    """
@external
def foo():
    x = 5
    x: int128 = 0
    """,
    """
@external
def foo():
    bork = zork
    """,
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_undeclared_def_exception(bad_code):
    with pytest.raises(UndeclaredDefinition):
        compiler.compile_code(bad_code)



# primitive types which occupy one word, like ints and addresses

from decimal import Decimal
from functools import cached_property
from typing import Any, Tuple, Union

from vyper import ast as vy_ast
from vyper.abi_types import ABI_Address, ABI_Bool, ABI_BytesM, ABI_GIntM, ABIType
from vyper.exceptions import (
    CompilerPanic,
    InvalidLiteral,
    InvalidOperation,
    OverflowException,
    VyperException,
)
from vyper.utils import checksum_encode, int_bounds, is_checksum_encoded

from .base import VyperType
from .bytestrings import BytesT


class _PrimT(VyperType):
    _is_prim_word = True
    _equality_attrs: tuple = ()
    _as_hashmap_key = True
    _as_array = True


# should inherit from uint8?
class BoolT(_PrimT):
    _id = "bool"
    _valid_literal = (vy_ast.NameConstant,)

    def validate_boolean_op(self, node: vy_ast.BoolOp) -> None:
        return

    def validate_numeric_op(
        self, node: Union[vy_ast.UnaryOp, vy_ast.BinOp, vy_ast.AugAssign]
    ) -> None:
        if not isinstance(node.op, vy_ast.Not):
            self._raise_invalid_op(node)

    @property
    def abi_type(self) -> ABIType:
        return ABI_Bool()

    def validate_literal(self, node: vy_ast.Constant) -> None:
        super().validate_literal(node)
        if node.value is None:
            raise InvalidLiteral("Invalid literal for type 'bool'", node)


RANGE_1_32 = list(range(1, 33))


# one-word bytesM with m possible bytes set, e.g. bytes1..bytes32
class BytesM_T(_PrimT):
    typeclass = "bytes_m"

    _valid_literal = (vy_ast.Hex,)

    _equality_attrs = ("m",)

    def __init__(self, m):
        super().__init__()
        self.m: int = m

    @property
    def _id(self):
        return f"bytes{self.m}"

    @property
    def m_bits(self):
        return self.m * 8

    # convenience for backwards API compat
    @property
    def length(self):
        return self.m

    @property
    def abi_type(self) -> ABIType:
        return ABI_BytesM(self.m)

    @classmethod
    def all(cls) -> Tuple["BytesM_T", ...]:
        return tuple(cls(m) for m in RANGE_1_32)

    def validate_literal(self, node: vy_ast.Constant) -> None:
        super().validate_literal(node)

        assert isinstance(node, vy_ast.Hex)  # keep mypy happy

        val = node.value

        if node.n_bytes != self.m:
            raise InvalidLiteral(f"Invalid literal for type {self}", node)

        nibbles = val[2:]  # strip leading 0x
        if nibbles not in (nibbles.lower(), nibbles.upper()):
            raise InvalidLiteral(f"Cannot mix uppercase and lowercase for {self} literal", node)

    def compare_type(self, other: VyperType) -> bool:
        if not super().compare_type(other):
            return False
        assert isinstance(other, BytesM_T)

        return self.m == other.m


class NumericT(_PrimT):
    _is_signed: bool
    _bits: int
    _invalid_ops: tuple

    # the type this can assume in the AST
    ast_type: type

    @property
    def ast_bounds(self):
        raise NotImplementedError("should be overridden!")

    # get the integer bounds on IR values of this type.
    # note the distinction for decimals: ast_bounds will return a Decimal,
    # int_bounds will return the fully expanded int range.
    @cached_property
    def int_bounds(self) -> Tuple[int, int]:
        return int_bounds(signed=self.is_signed, bits=self.bits)

    @cached_property
    def bits(self) -> int:
        return self._bits

    @cached_property
    def is_signed(self) -> bool:
        return self._is_signed

    def validate_literal(self, node: vy_ast.Constant) -> None:
        super().validate_literal(node)
        lower, upper = self.ast_bounds
        if node.value < lower:
            raise OverflowException(f"Value is below lower bound for given type ({lower})", node)
        if node.value > upper:
            raise OverflowException(f"Value exceeds upper bound for given type ({upper})", node)

    def validate_numeric_op(
        self, node: Union[vy_ast.UnaryOp, vy_ast.BinOp, vy_ast.AugAssign]
    ) -> None:
        if isinstance(node.op, self._invalid_ops):
            self._raise_invalid_op(node)

        def _get_lr():
            if isinstance(node, vy_ast.BinOp):
                return node.left.reduced(), node.right.reduced()
            elif isinstance(node, vy_ast.AugAssign):
                return node.target.reduced(), node.value.reduced()
            else:
                raise CompilerPanic(f"Unexpected node type for numeric op: {type(node).__name__}")

        if isinstance(node.op, (vy_ast.LShift, vy_ast.RShift)):
            if self._bits != 256:
                raise InvalidOperation(
                    f"Cannot perform {node.op.description} on non-int256/uint256 type!", node
                )

        if isinstance(node.op, vy_ast.Pow):
            left, right = _get_lr()

            value_bits = self._bits - (1 if self._is_signed else 0)

            # TODO double check: this code seems duplicated with constant eval
            # constant folding ensures one of `(left, right)` is never a literal
            if isinstance(left, vy_ast.Int):
                if left.value >= 2**value_bits:
                    raise OverflowException(
                        "Base is too large, calculation will always overflow", left
                    )
                elif left.value < -(2**value_bits):
                    raise OverflowException(
                        "Base is too small, calculation will always underflow", left
                    )
            elif isinstance(right, vy_ast.Int):
                if right.value < 0:
                    raise InvalidOperation("Cannot calculate a negative power", right)
                if right.value > value_bits:
                    raise OverflowException(
                        "Power is too large, calculation will always overflow", right
                    )
            else:
                msg = (
                    "Cannot apply an overflow check on exponentiation when both "
                    "the base and power are unknown at compile-time."
                )
                if not self._is_signed:
                    msg = (
                        f"{msg} To perform this operation without an overflow check, use "
                        f"`pow_mod256({left.node_source_code}, {right.node_source_code})`"
                    )
                raise InvalidOperation(msg, node)

    def validate_comparator(self, node: vy_ast.Compare) -> None:
        # all comparators are valid on numeric types
        return


def _add_div_hint(node, e):
    if isinstance(node.op, vy_ast.Div):
        suggested = vy_ast.FloorDiv._pretty
    elif isinstance(node.op, vy_ast.FloorDiv):
        suggested = vy_ast.Div._pretty
    else:
        return e

    if isinstance(node, vy_ast.BinOp):
        e._hint = f"did you mean `{node.left.node_source_code} "
        e._hint += f"{suggested} {node.right.node_source_code}`?"
    elif isinstance(node, vy_ast.AugAssign):
        e._hint = f"did you mean `{node.target.node_source_code} "
        e._hint += f"{suggested}= {node.value.node_source_code}`?"

    return e


class IntegerT(NumericT):
    """
    General integer type. All signed and unsigned ints from uint8 thru int256

    Attributes
    ----------
    bits : int
        Number of bits the value occupies in memory
    is_signed : bool
        Is the value signed?
    """

    typeclass = "integer"

    _valid_literal = (vy_ast.Int,)
    _equality_attrs = ("is_signed", "bits")

    ast_type = int

    def __init__(self, is_signed, bits):
        super().__init__()
        self._is_signed = is_signed
        self._bits = bits

    @cached_property
    def _id(self):
        u = "u" if not self.is_signed else ""
        return f"{u}int{self.bits}"

    @cached_property
    def ast_bounds(self) -> Tuple[int, int]:
        return int_bounds(self.is_signed, self.bits)

    @cached_property
    def _invalid_ops(self):
        invalid_ops = (vy_ast.Not, vy_ast.Div)
        if not self.is_signed:
            return invalid_ops + (vy_ast.USub,)
        return invalid_ops

    def validate_numeric_op(self, node) -> None:
        try:
            super().validate_numeric_op(node)
        except VyperException as e:
            raise _add_div_hint(node, e) from None

    @classmethod
    # TODO maybe cache these three classmethods
    def signeds(cls) -> Tuple["IntegerT", ...]:
        return tuple(cls(is_signed=True, bits=i * 8) for i in RANGE_1_32)

    @classmethod
    def unsigneds(cls) -> Tuple["IntegerT", ...]:
        return tuple(cls(is_signed=False, bits=i * 8) for i in RANGE_1_32)

    @classmethod
    def all(cls) -> Tuple["IntegerT", ...]:
        return cls.signeds() + cls.unsigneds()

    @cached_property
    def abi_type(self) -> ABIType:
        return ABI_GIntM(self.bits, self.is_signed)

    def compare_type(self, other: VyperType) -> bool:
        # this function is performance sensitive
        # originally:
        # if not super().compare_type(other):
        #    return False
        # return self.is_signed == other.is_signed and self.bits == other.bits

        return (  # noqa: E721
            self.__class__ == other.__class__
            and self.is_signed == other.is_signed  # type: ignore
            and self.bits == other.bits  # type: ignore
        )


# helper function for readability.
# returns a uint<N> type.
def UINT(bits):
    return IntegerT(False, bits)


# helper function for readability.
# returns an int<N> type.
def SINT(bits):
    return IntegerT(True, bits)


class DecimalT(NumericT):
    typeclass = "decimal"

    _bits = 168  # TODO generalize
    _decimal_places = 10  # TODO generalize
    _id = "decimal"
    _is_signed = True
    _invalid_ops = (
        vy_ast.Pow,
        vy_ast.FloorDiv,
        vy_ast.BitAnd,
        vy_ast.BitOr,
        vy_ast.BitXor,
        vy_ast.Not,
    )
    _valid_literal = (vy_ast.Decimal,)

    _equality_attrs = ("_bits", "_decimal_places")

    ast_type = Decimal

    def validate_numeric_op(self, node) -> None:
        try:
            super().validate_numeric_op(node)
        except VyperException as e:
            raise _add_div_hint(node, e) from None

    @cached_property
    def abi_type(self) -> ABIType:
        return ABI_GIntM(self._bits, self._is_signed)

    @cached_property
    def decimals(self) -> int:
        # Alias for API compatibility with codegen
        return self._decimal_places

    @cached_property
    def divisor(self) -> int:
        return 10**self.decimals

    @cached_property
    def epsilon(self) -> Decimal:
        return 1 / Decimal(self.divisor)

    @cached_property
    def ast_bounds(self) -> Tuple[Decimal, Decimal]:
        return self.decimal_bounds

    @cached_property
    def decimal_bounds(self) -> Tuple[Decimal, Decimal]:
        lo, hi = int_bounds(signed=self.is_signed, bits=self.bits)
        DIVISOR = Decimal(self.divisor)
        return lo / DIVISOR, hi / DIVISOR

    def to_abi_arg(self, name: str = "") -> dict[str, Any]:
        ret = super().to_abi_arg(name)
        ret["internalType"] = repr(self)
        return ret


# maybe this even deserves its own module, address.py
# should inherit from uint160?
class AddressT(_PrimT):
    _id = "address"
    _valid_literal = (vy_ast.Hex,)
    _type_members = {
        "balance": UINT(256),
        "codehash": BytesM_T(32),
        "codesize": UINT(256),
        "is_contract": BoolT(),
        "code": BytesT(),
    }

    @cached_property
    def abi_type(self) -> ABIType:
        return ABI_Address()

    def validate_literal(self, node: vy_ast.Constant) -> None:
        super().validate_literal(node)
        assert isinstance(node, vy_ast.Hex)  # keep mypy happy
        if node.n_bytes != 20:
            raise InvalidLiteral(f"Invalid address. Expected 20 bytes, got {node.n_bytes}.", node)

        addr = node.value
        if not is_checksum_encoded(addr):
            raise InvalidLiteral(
                "Address checksum mismatch. If you are sure this is the right "
                f"address, the correct checksummed form is: {checksum_encode(addr)}",
                node,
            )


# type for "self"
# refactoring note: it might be best for this to be a ModuleT actually
class SelfT(AddressT):
    _id = "self"

    def compare_type(self, other):
        # compares true to AddressT
        return isinstance(other, type(self)) or isinstance(self, type(other))



# validation utils for ast
from typing import Optional, Union

from vyper.ast import nodes as vy_ast
from vyper.exceptions import ArgumentException, CompilerPanic, StructureException


# TODO this really belongs in vyper/semantics/validation/utils
def validate_call_args(
    node: vy_ast.Call, arg_count: Union[int, tuple], kwargs: Optional[list] = None
) -> None:
    """
    Validate positional and keyword arguments of a Call node.

    This function does not handle type checking of arguments, it only checks
    correctness of the number of arguments given and keyword names.

    Arguments
    ---------
    node : Call
        Vyper ast Call node to be validated.
    arg_count : int | tuple
        The required number of positional arguments. When given as a tuple the
        value is interpreted as the minimum and maximum number of arguments.
        If arg_count[1] is -1, it is interpreted to mean that there is no
        limit to the number of arguments (ex. when the function has a
        variable number of arguments).
    kwargs : list, optional
        A list of valid keyword arguments. When arg_count is a tuple and the
        number of positional arguments exceeds the minimum, the excess values are
        considered to fill the first values on this list.

    Returns
    -------
        None. Raises an exception when the arguments are invalid.
    """

    if not isinstance(node, vy_ast.Call):
        raise StructureException("Expected Call", node)
    if not isinstance(arg_count, (int, tuple)):
        raise CompilerPanic(f"Invalid type for arg_count: {type(arg_count).__name__}")

    if isinstance(arg_count, tuple):
        if arg_count[1] == -1:
            # -1 is sentinel which means we have varargs.
            # set arg_count[1] to some large number that we
            # would never see in practice
            arg_count = (arg_count[0], 2**64)

        if arg_count[0] == arg_count[1]:
            arg_count = arg_count[0]

    if isinstance(node.func, vy_ast.Attribute):
        msg = f" for call to '{node.func.attr}'"
    elif isinstance(node.func, vy_ast.Name):
        msg = f" for call to '{node.func.id}'"
    else:
        raise CompilerPanic("Unreachable")

    if isinstance(arg_count, int) and len(node.args) != arg_count:
        if not node.args:
            exc_node = node
        elif len(node.args) < arg_count:
            exc_node = node.args[-1]
        else:
            exc_node = node.args[arg_count]
        raise ArgumentException(
            f"Invalid argument count{msg}: expected {arg_count}, got {len(node.args)}", exc_node
        )

    if isinstance(arg_count, tuple) and not arg_count[0] <= len(node.args) <= arg_count[1]:
        if not node.args:
            exc_node = node
        elif len(node.args) < arg_count[0]:
            exc_node = node.args[-1]
        else:
            exc_node = node.args[arg_count[1]]
        raise ArgumentException(
            f"Invalid argument count{msg}: expected {arg_count[0]} "
            f"to {arg_count[1]}, got {len(node.args)}",
            exc_node,
        )

    if kwargs is None:
        if node.keywords:
            raise ArgumentException("Keyword arguments are not accepted here", node.keywords[0])
        return

    kwargs_seen = set()
    for key in node.keywords:
        if key.arg is None:
            raise StructureException("Use of **kwargs is not supported", key.value)
        if key.arg not in kwargs:
            raise ArgumentException(f"Invalid keyword argument '{key.arg}'", key)
        if key.arg in kwargs_seen:
            raise ArgumentException(f"Duplicate keyword argument '{key.arg}'", key)
        kwargs_seen.add(key.arg)



import pytest

from tests.utils import ZERO_ADDRESS

EXPIRY = 16


@pytest.fixture(scope="module")
def auction_start(env):
    return env.timestamp + 1


@pytest.fixture(scope="module")
def auction_contract(env, get_contract, auction_start):
    with open("examples/auctions/simple_open_auction.vy") as f:
        contract_code = f.read()

    for acc in env.accounts[:5]:
        env.set_balance(acc, 10**20)

    env.timestamp += 1  # make sure auction has started
    return get_contract(contract_code, *[env.accounts[0], auction_start, EXPIRY])


def test_initial_state(env, auction_contract, auction_start):
    # Check beneficiary is correct
    assert auction_contract.beneficiary() == env.accounts[0]
    # Check start time is `auction_start`
    assert auction_contract.auctionStart() == auction_start
    # Check time difference between start time and end time is EXPIRY
    assert auction_contract.auctionEnd() == auction_contract.auctionStart() + EXPIRY
    # Check auction has not ended
    assert auction_contract.ended() is False
    # Check highest bidder is empty
    assert auction_contract.highestBidder() == ZERO_ADDRESS
    # Check highest bid is 0
    assert auction_contract.highestBid() == 0
    # Check end time is more than current block timestamp
    assert auction_contract.auctionEnd() >= env.timestamp


def test_bid(env, auction_contract, tx_failed):
    k1, k2, k3, k4, k5 = env.accounts[:5]

    # Bidder cannot bid 0
    with tx_failed():
        auction_contract.bid(value=0, sender=k1)

    # Bidder can bid
    auction_contract.bid(value=1, sender=k1)
    # Check that highest bidder and highest bid have changed accordingly
    assert auction_contract.highestBidder() == k1
    assert auction_contract.highestBid() == 1
    # Bidder bid cannot equal current highest bid
    with tx_failed():
        auction_contract.bid(value=1, sender=k1)
    # Higher bid can replace current highest bid
    auction_contract.bid(value=2, sender=k2)
    # Check that highest bidder and highest bid have changed accordingly
    assert auction_contract.highestBidder() == k2
    assert auction_contract.highestBid() == 2
    # Multiple bidders can bid
    auction_contract.bid(value=3, sender=k3)
    auction_contract.bid(value=4, sender=k4)
    auction_contract.bid(value=5, sender=k5)
    # Check that highest bidder and highest bid have changed accordingly
    assert auction_contract.highestBidder() == k5
    assert auction_contract.highestBid() == 5
    auction_contract.bid(value=1 * 10**10, sender=k1)
    pending_return_before_outbid = auction_contract.pendingReturns(k1)
    auction_contract.bid(value=2 * 10**10, sender=k2)
    pending_return_after_outbid = auction_contract.pendingReturns(k1)
    # Account has a greater pending return balance after being outbid
    assert pending_return_after_outbid > pending_return_before_outbid

    balance_before_withdrawal = env.get_balance(k1)
    auction_contract.withdraw(sender=k1)
    balance_after_withdrawal = env.get_balance(k1)
    # Balance increases after withdrawal
    assert balance_after_withdrawal > balance_before_withdrawal
    # Pending return balance is reset to 0
    assert auction_contract.pendingReturns(k1) == 0


def test_end_auction(env, auction_contract, tx_failed):
    k1, k2, k3, k4, k5 = env.accounts[:5]

    # Fails if auction end time has not been reached
    with tx_failed():
        auction_contract.endAuction()

    auction_contract.bid(value=1 * 10**10, sender=k2)
    # Move block timestamp forward to reach auction end time
    env.timestamp += EXPIRY
    balance_before_end = env.get_balance(k1)
    auction_contract.endAuction(sender=k2)
    balance_after_end = env.get_balance(k1)
    # Beneficiary receives the highest bid
    assert balance_after_end == balance_before_end + 1 * 10**10
    # Bidder cannot bid after auction end time has been reached
    with tx_failed():
        auction_contract.bid(value=10, sender=k1)
    # Auction cannot be ended twice
    with tx_failed():
        auction_contract.endAuction()



import os
from pathlib import Path, PurePath
from typing import Any, Optional

import vyper.builtins.interfaces
from vyper import ast as vy_ast
from vyper.compiler.input_bundle import (
    ABIInput,
    CompilerInput,
    FileInput,
    FilesystemInputBundle,
    InputBundle,
    PathLike,
)
from vyper.evm.opcodes import version_check
from vyper.exceptions import (
    BorrowException,
    CallViolation,
    CompilerPanic,
    DuplicateImport,
    EvmVersionException,
    ExceptionList,
    ImmutableViolation,
    InitializerException,
    InterfaceViolation,
    InvalidLiteral,
    InvalidType,
    ModuleNotFound,
    StateAccessViolation,
    StructureException,
    UndeclaredDefinition,
    VyperException,
    tag_exceptions,
)
from vyper.semantics.analysis.base import (
    ExportsInfo,
    ImportInfo,
    InitializesInfo,
    Modifiability,
    ModuleInfo,
    ModuleOwnership,
    UsesInfo,
    VarInfo,
)
from vyper.semantics.analysis.common import VyperNodeVisitorBase
from vyper.semantics.analysis.constant_folding import constant_fold
from vyper.semantics.analysis.getters import generate_public_variable_getters
from vyper.semantics.analysis.import_graph import ImportGraph
from vyper.semantics.analysis.local import ExprVisitor, analyze_functions, check_module_uses
from vyper.semantics.analysis.utils import (
    check_modifiability,
    get_exact_type_from_node,
    get_expr_info,
)
from vyper.semantics.data_locations import DataLocation
from vyper.semantics.namespace import Namespace, get_namespace, override_global_namespace
from vyper.semantics.types import EventT, FlagT, InterfaceT, StructT
from vyper.semantics.types.function import ContractFunctionT
from vyper.semantics.types.module import ModuleT
from vyper.semantics.types.utils import type_from_annotation
from vyper.utils import OrderedSet


def analyze_module(
    module_ast: vy_ast.Module,
    input_bundle: InputBundle,
    import_graph: ImportGraph = None,
    is_interface: bool = False,
) -> ModuleT:
    """
    Analyze a Vyper module AST node, recursively analyze all its imports,
    add all module-level objects to the namespace, type-check/validate
    semantics and annotate with type and analysis info
    """
    if import_graph is None:
        import_graph = ImportGraph()

    return _analyze_module_r(module_ast, input_bundle, import_graph, is_interface)


def _analyze_module_r(
    module_ast: vy_ast.Module,
    input_bundle: InputBundle,
    import_graph: ImportGraph,
    is_interface: bool = False,
):
    if "type" in module_ast._metadata:
        # we don't need to analyse again, skip out
        assert isinstance(module_ast._metadata["type"], ModuleT)
        return module_ast._metadata["type"]

    # validate semantics and annotate AST with type/semantics information
    namespace = get_namespace()

    with namespace.enter_scope(), import_graph.enter_path(module_ast):
        analyzer = ModuleAnalyzer(module_ast, input_bundle, namespace, import_graph, is_interface)
        analyzer.analyze_module_body()

        _analyze_call_graph(module_ast)
        generate_public_variable_getters(module_ast)

        ret = ModuleT(module_ast)
        module_ast._metadata["type"] = ret

        # if this is an interface, the function is already validated
        # in `ContractFunction.from_vyi()`
        if not is_interface:
            analyze_functions(module_ast)
            analyzer.validate_initialized_modules()
            analyzer.validate_used_modules()

    return ret


def _analyze_call_graph(module_ast: vy_ast.Module):
    # get list of internal function calls made by each function
    # CMC 2024-02-03 note: this could be cleaner in analysis/local.py
    function_defs = module_ast.get_children(vy_ast.FunctionDef)

    for func in function_defs:
        fn_t = func._metadata["func_type"]
        assert len(fn_t.called_functions) == 0
        fn_t.called_functions = OrderedSet()

        function_calls = func.get_descendants(vy_ast.Call)

        for call in function_calls:
            try:
                call_t = get_exact_type_from_node(call.func)
            except VyperException:
                # there is a problem getting the call type. this might be
                # an issue, but it will be handled properly later. right now
                # we just want to be able to construct the call graph.
                continue

            if isinstance(call_t, ContractFunctionT) and (
                call_t.is_internal or call_t.is_constructor
            ):
                fn_t.called_functions.add(call_t)

    for func in function_defs:
        fn_t = func._metadata["func_type"]

        # compute reachable set and validate the call graph
        _compute_reachable_set(fn_t)


# compute reachable set and validate the call graph (detect cycles)
def _compute_reachable_set(fn_t: ContractFunctionT, path: list[ContractFunctionT] = None) -> None:
    path = path or []

    path.append(fn_t)
    root = path[0]

    for g in fn_t.called_functions:
        if g in fn_t.reachable_internal_functions:
            # already seen
            continue

        if g == root:
            message = " -> ".join([f.name for f in path])
            raise CallViolation(f"Contract contains cyclic function call: {message}")

        _compute_reachable_set(g, path=path)

        g_reachable = g.reachable_internal_functions
        assert fn_t not in g_reachable  # sanity check
        fn_t.reachable_internal_functions.update(g_reachable)

        fn_t.reachable_internal_functions.add(g)

    path.pop()


class ModuleAnalyzer(VyperNodeVisitorBase):
    scope_name = "module"

    def __init__(
        self,
        module_node: vy_ast.Module,
        input_bundle: InputBundle,
        namespace: Namespace,
        import_graph: ImportGraph,
        is_interface: bool = False,
    ) -> None:
        self.ast = module_node
        self.input_bundle = input_bundle
        self.namespace = namespace
        self._import_graph = import_graph
        self.is_interface = is_interface

        # keep track of imported modules to prevent duplicate imports
        self._imported_modules: dict[PurePath, vy_ast.VyperNode] = {}

        # keep track of exported functions to prevent duplicate exports
        self._exposed_functions: dict[ContractFunctionT, vy_ast.VyperNode] = {}

        self._events: list[EventT] = []

        self.module_t: Optional[ModuleT] = None

        # ast cache, hitchhike onto the input_bundle object
        if not hasattr(self.input_bundle._cache, "_ast_of"):
            self.input_bundle._cache._ast_of: dict[int, vy_ast.Module] = {}  # type: ignore

    def analyze_module_body(self):
        # generate a `ModuleT` from the top-level node
        # note: also validates unique method ids

        assert "type" not in self.ast._metadata

        self._to_visit = self.ast.body.copy()

        # handle imports; mutates `self._imported_modules`
        self._visit_nodes_linear((vy_ast.Import, vy_ast.ImportFrom))

        # we can resolve constants as soon as imports are handled.
        constant_fold(self.ast)

        # handle ownership decls, mutate ModuleInfo.ownership
        self._visit_nodes_linear((vy_ast.UsesDecl, vy_ast.InitializesDecl))

        # handle some node types using a dependency resolution routine
        # which loops, swallowing exceptions until all nodes are processed
        type_decls = (vy_ast.FlagDef, vy_ast.StructDef, vy_ast.InterfaceDef, vy_ast.EventDef)
        self._visit_nodes_looping(type_decls)

        # handle functions
        # run before exports for exception handling priority
        self._visit_nodes_looping((vy_ast.VariableDecl, vy_ast.FunctionDef))

        # mutates _exposed_functions
        self._visit_nodes_linear(vy_ast.ExportsDecl)

        # handle implements last, after all functions are handled
        self._visit_nodes_linear(vy_ast.ImplementsDecl)

        # we are done! any remaining nodes should raise errors; visit
        # them to trip the exception.
        for n in self._to_visit:
            self.visit(n)

        # attach namespace to the module for downstream use.
        _ns = Namespace()
        # note that we don't just copy the namespace because
        # there are constructor issues.
        _ns.update({k: self.namespace[k] for k in self.namespace._scopes[-1]})  # type: ignore
        self.ast._metadata["namespace"] = _ns

    def _visit_nodes_linear(self, node_type):
        for node in self._to_visit.copy():
            if not isinstance(node, node_type):
                continue
            self.visit(node)
            self._to_visit.remove(node)

    # visit nodes which may have dependencies on each other
    def _visit_nodes_looping(self, node_type):
        nodes = [n for n in self._to_visit if isinstance(n, node_type)]

        # keep trying to process all the nodes until we finish or can
        # no longer progress. this makes it so we don't need to
        # calculate a dependency tree between top-level items.
        # note that the nodes processed here should not mutate ModuleAnalyzer
        # state, otherwise ModuleAnalyzer state can end up invalid!
        while len(nodes) > 0:
            count = len(nodes)
            err_list = ExceptionList()
            for node in nodes.copy():
                try:
                    self.visit(node)
                    nodes.remove(node)
                    self._to_visit.remove(node)
                except (InvalidLiteral, InvalidType) as e:
                    # these exceptions cannot be caused by another statement
                    # not yet being parsed, so we raise them immediately
                    raise e from None
                except VyperException as e:
                    err_list.append(e)

            # Only raise if no nodes were successfully processed. This allows
            # module level logic to parse regardless of the ordering of code
            # elements.
            if count == len(nodes):
                err_list.raise_if_not_empty()

    def validate_used_modules(self):
        # check all `uses:` modules are actually used
        should_use = {}

        module_t = self.ast._metadata["type"]
        uses_decls = module_t.uses_decls
        for decl in uses_decls:
            info = decl._metadata["uses_info"]
            for m in info.used_modules:
                should_use[m.module_t] = (m, info)

        initialized_modules = {t.module_info.module_t: t for t in module_t.initialized_modules}

        all_used_modules = OrderedSet()

        for f in module_t.functions.values():
            for u in f.get_used_modules():
                all_used_modules.add(u.module_t)

        for decl in module_t.exports_decls:
            info = decl._metadata["exports_info"]
            all_used_modules.update([u.module_t for u in info.used_modules])

        for used_module in all_used_modules:
            if used_module in initialized_modules:
                continue

            if used_module in should_use:
                del should_use[used_module]

        if len(should_use) > 0:
            err_list = ExceptionList()
            for used_module_info, uses_info in should_use.values():
                msg = f"`{used_module_info.alias}` is declared as used, but "
                msg += f"its state is not actually used in {module_t}!"
                hint = f"delete `uses: {used_module_info.alias}`"
                err_list.append(BorrowException(msg, uses_info.node, hint=hint))

            err_list.raise_if_not_empty()

    def validate_initialized_modules(self):
        # check all `initializes:` modules have `__init__()` called exactly once
        module_t = self.ast._metadata["type"]
        should_initialize = {t.module_info.module_t: t for t in module_t.initialized_modules}
        # don't call `__init__()` for modules which don't have
        # `__init__()` function
        for m in should_initialize.copy():
            for f in m.functions.values():
                if f.is_constructor:
                    break
            else:
                del should_initialize[m]

        init_calls = []
        for f in self.ast.get_children(vy_ast.FunctionDef):
            if f._metadata["func_type"].is_constructor:
                init_calls = f.get_descendants(vy_ast.Call)
                break

        seen_initializers = {}
        for call_node in init_calls:
            expr_info = call_node.func._expr_info
            if expr_info is None:
                # this can happen for range() calls; CMC 2024-02-05 try to
                # refactor so that range() is properly tagged.
                continue

            call_t = call_node.func._expr_info.typ

            if not isinstance(call_t, ContractFunctionT):
                continue

            if not call_t.is_constructor:
                continue

            # XXX: check this works as expected for nested attributes
            initialized_module = call_node.func.value._expr_info.module_info

            if initialized_module.module_t in seen_initializers:
                seen_location = seen_initializers[initialized_module.module_t]
                msg = f"tried to initialize `{initialized_module.alias}`, "
                msg += "but its __init__() function was already called!"
                raise InitializerException(msg, call_node.func, seen_location)

            if initialized_module.module_t not in should_initialize:
                msg = f"tried to initialize `{initialized_module.alias}`, "
                msg += "but it is not in initializer list!"
                hint = f"add `initializes: {initialized_module.alias}` "
                hint += "as a top-level statement to your contract"
                raise InitializerException(msg, call_node.func, hint=hint)

            del should_initialize[initialized_module.module_t]
            seen_initializers[initialized_module.module_t] = call_node.func

        if len(should_initialize) > 0:
            err_list = ExceptionList()
            for s in should_initialize.values():
                msg = "not initialized!"
                hint = f"add `{s.module_info.alias}.__init__()` to "
                hint += "your `__init__()` function"

                # grab the init function AST node for error message
                # (it could be None, it's ok since it's just for diagnostics)
                init_func_node = None
                if module_t.init_function:
                    init_func_node = module_t.init_function.decl_node
                err_list.append(InitializerException(msg, init_func_node, s.node, hint=hint))

            err_list.raise_if_not_empty()

    def _ast_from_file(self, file: FileInput) -> vy_ast.Module:
        # cache ast if we have seen it before.
        # this gives us the additional property of object equality on
        # two ASTs produced from the same source
        ast_of = self.input_bundle._cache._ast_of
        if file.source_id not in ast_of:
            ast_of[file.source_id] = _parse_ast(file)

        return ast_of[file.source_id]

    def visit_ImplementsDecl(self, node):
        type_ = type_from_annotation(node.annotation)

        if not isinstance(type_, InterfaceT):
            msg = "Not an interface!"
            hint = None
            if isinstance(type_, ModuleT):
                path = type_._module.path
                msg += " (Since vyper v0.4.0, interface files are required"
                msg += " to have a .vyi suffix.)"
                hint = f"try renaming `{path}` to `{path}i`"
            raise StructureException(msg, node.annotation, hint=hint)

        # grab exposed functions
        funcs = self._exposed_functions
        type_.validate_implements(node, funcs)

        node._metadata["interface_type"] = type_

    def visit_UsesDecl(self, node):
        # TODO: check duplicate uses declarations, e.g.
        # uses: x
        # ...
        # uses: x
        items = vy_ast.as_tuple(node.annotation)

        used_modules = []

        for item in items:
            module_info = get_expr_info(item).module_info
            if module_info is None:
                raise StructureException("not a valid module!", item)

            # note: try to refactor - not a huge fan of mutating the
            # ModuleInfo after it's constructed
            module_info.set_ownership(ModuleOwnership.USES, item)

            used_modules.append(module_info)

        node._metadata["uses_info"] = UsesInfo(used_modules, node)

    def visit_InitializesDecl(self, node):
        annotation = node.annotation

        dependencies_ast = ()
        module_ref = annotation
        if isinstance(module_ref, vy_ast.Subscript):
            dependencies_ast = vy_ast.as_tuple(module_ref.slice)
            module_ref = module_ref.value

        # postcondition of InitializesDecl.validates()
        assert isinstance(module_ref, (vy_ast.Name, vy_ast.Attribute))

        module_info = get_expr_info(module_ref).module_info
        if module_info is None:
            raise StructureException("Not a module!", module_ref)

        used_modules = {i.module_t: i for i in module_info.module_t.used_modules}

        dependencies = []
        for named_expr in dependencies_ast:
            assert isinstance(named_expr, vy_ast.NamedExpr)

            rhs_module = get_expr_info(named_expr.value).module_info

            with module_info.module_node.namespace():
                # lhs of the named_expr is evaluated in the namespace of the
                # initialized module!
                try:
                    lhs_module = get_expr_info(named_expr.target).module_info
                except VyperException as e:
                    # try to report a common problem - user names the module in
                    # the current namespace instead of the initialized module
                    # namespace.

                    # search for the module in the initialized module
                    found_module = module_info.module_t.find_module_info(rhs_module.module_t)
                    if found_module is not None:
                        msg = f"unknown module `{named_expr.target.id}`"
                        hint = f"did you mean `{found_module.alias} := {rhs_module.alias}`?"
                        raise UndeclaredDefinition(msg, named_expr.target, hint=hint)

                    raise e from None

            if lhs_module.module_t != rhs_module.module_t:
                raise StructureException(
                    f"{lhs_module.alias} is not {rhs_module.alias}!", named_expr
                )
            dependencies.append(lhs_module)

            if lhs_module.module_t not in used_modules:
                raise InitializerException(
                    f"`{module_info.alias}` is initialized with `{lhs_module.alias}`, "
                    f"but `{module_info.alias}` does not use `{lhs_module.alias}`!",
                    named_expr,
                )

            del used_modules[lhs_module.module_t]

        if len(used_modules) > 0:
            item = next(iter(used_modules.values()))  # just pick one
            msg = f"`{module_info.alias}` uses `{item.alias}`, but it is not "
            msg += f"initialized with `{item.alias}`"

            lhs = item.alias
            rhs = None
            # find the alias of the uninitialized module in this contract
            # to fill out the error message with.
            for k, v in self.namespace.items():
                if isinstance(v, ModuleInfo) and v.module_t == item.module_t:
                    rhs = k
                    break

            if rhs is None:
                hint = f"try importing {item.alias} first"
            elif not isinstance(annotation, vy_ast.Subscript):
                # it's `initializes: foo` instead of `initializes: foo[...]`
                hint = f"did you mean {module_ref.id}[{lhs} := {rhs}]?"
            else:
                hint = f"add `{lhs} := {rhs}` to its initializer list"
            raise InitializerException(msg, node, hint=hint)

        # note: try to refactor. not a huge fan of mutating the
        # ModuleInfo after it's constructed
        module_info.set_ownership(ModuleOwnership.INITIALIZES, node)
        node._metadata["initializes_info"] = InitializesInfo(module_info, dependencies, node)

    def visit_ExportsDecl(self, node):
        items = vy_ast.as_tuple(node.annotation)
        exported_funcs = []
        used_modules = OrderedSet()

        # CMC 2024-04-13 TODO: reduce nesting in this function

        for item in items:
            # set is_callable=True to give better error messages for imported
            # types, e.g. exports: some_module.MyEvent
            info = get_expr_info(item, is_callable=True)

            if info.var_info is not None:
                decl = info.var_info.decl_node
                if not info.var_info.is_public:
                    raise StructureException("not a public variable!", decl, item)
                funcs = [decl._expanded_getter._metadata["func_type"]]
            elif isinstance(info.typ, ContractFunctionT):
                # regular function
                funcs = [info.typ]
            elif isinstance(info.typ, InterfaceT):
                if not isinstance(item, vy_ast.Attribute):
                    raise StructureException(
                        "invalid export",
                        hint="exports should look like <module>.<function | interface>",
                    )

                module_info = get_expr_info(item.value).module_info
                if module_info is None:
                    raise StructureException("not a valid module!", item.value)

                if info.typ not in module_info.typ.implemented_interfaces:
                    iface_str = item.node_source_code
                    module_str = item.value.node_source_code
                    msg = f"requested `{iface_str}` but `{module_str}`"
                    msg += f" does not implement `{iface_str}`!"
                    raise InterfaceViolation(msg, item)

                module_exposed_fns = {fn.name: fn for fn in module_info.typ.exposed_functions}
                # find the specific implementation of the function in the module
                funcs = [
                    module_exposed_fns[fn.name]
                    for fn in info.typ.functions.values()
                    if fn.is_external
                ]
            else:
                raise StructureException(
                    f"not a function or interface: `{info.typ}`", info.typ.decl_node, item
                )

            for func_t in funcs:
                if not func_t.is_external:
                    raise StructureException(
                        "can't export non-external functions!", func_t.decl_node, item
                    )

                self._add_exposed_function(func_t, item, relax=False)
                with tag_exceptions(item):  # tag exceptions with specific item
                    self._self_t.typ.add_member(func_t.name, func_t)

                    exported_funcs.append(func_t)

                    # check module uses
                    if func_t.uses_state():
                        module_info = check_module_uses(item)

                        # guaranteed by above checks:
                        assert module_info is not None

                        used_modules.add(module_info)

        node._metadata["exports_info"] = ExportsInfo(exported_funcs, used_modules)

    @property
    def _self_t(self):
        return self.namespace["self"]

    def _add_exposed_function(self, func_t, node, relax=True):
        # call this before self._self_t.typ.add_member() for exception raising
        # priority
        if not relax and (prev_decl := self._exposed_functions.get(func_t)) is not None:
            raise StructureException("already exported!", node, prev_decl=prev_decl)

        self._exposed_functions[func_t] = node

    def visit_VariableDecl(self, node):
        # postcondition of VariableDecl.validate
        assert isinstance(node.target, vy_ast.Name)
        name = node.target.id

        if node.is_public:
            # generate function type and add to metadata
            # we need this when building the public getter
            func_t = ContractFunctionT.getter_from_VariableDecl(node)
            node._metadata["getter_type"] = func_t
            self._add_exposed_function(func_t, node)

        # TODO: move this check to local analysis
        if node.is_immutable:
            # mutability is checked automatically preventing assignment
            # outside of the constructor, here we just check a value is assigned,
            # not necessarily where
            assignments = self.ast.get_descendants(
                vy_ast.Assign, filters={"target.id": node.target.id}
            )
            if not assignments:
                # Special error message for common wrong usages via `self.<immutable name>`
                wrong_self_attribute = self.ast.get_descendants(
                    vy_ast.Attribute, {"value.id": "self", "attr": node.target.id}
                )
                message = (
                    "Immutable variables must be accessed without 'self'"
                    if len(wrong_self_attribute) > 0
                    else "Immutable definition requires an assignment in the constructor"
                )
                raise ImmutableViolation(message, node)

        data_loc = (
            DataLocation.CODE
            if node.is_immutable
            else DataLocation.UNSET
            if node.is_constant
            else DataLocation.TRANSIENT
            if node.is_transient
            else DataLocation.STORAGE
        )

        modifiability = (
            Modifiability.RUNTIME_CONSTANT
            if node.is_immutable
            else Modifiability.CONSTANT
            if node.is_constant
            else Modifiability.MODIFIABLE
        )

        type_ = type_from_annotation(node.annotation, data_loc)

        if node.is_transient and not version_check(begin="cancun"):
            raise EvmVersionException("`transient` is not available pre-cancun", node.annotation)

        var_info = VarInfo(
            type_,
            decl_node=node,
            location=data_loc,
            modifiability=modifiability,
            is_public=node.is_public,
        )
        node.target._metadata["varinfo"] = var_info  # TODO maybe put this in the global namespace
        node._metadata["type"] = type_

        def _finalize():
            # add the variable name to `self` namespace if the variable is either
            # 1. a public constant or immutable; or
            # 2. a storage variable, whether private or public
            if (node.is_constant or node.is_immutable) and not node.is_public:
                return

            self._self_t.typ.add_member(name, var_info)
            node.target._metadata["type"] = type_

        def _validate_self_namespace():
            # block globals if storage variable already exists
            self._self_t.typ._check_add_member(name)
            self.namespace[name] = var_info

        if node.is_constant:
            assert node.value is not None  # checked in VariableDecl.validate()
            ExprVisitor().visit(node.value, type_)  # performs validate_expected_type

            if not check_modifiability(node.value, Modifiability.CONSTANT):
                raise StateAccessViolation("Value must be a literal", node.value)

            _validate_self_namespace()

            return _finalize()

        assert node.value is None  # checked in VariableDecl.validate()
        if node.is_immutable:
            _validate_self_namespace()
            return _finalize()

        self.namespace.validate_assignment(name)

        return _finalize()

    def visit_FlagDef(self, node):
        obj = FlagT.from_FlagDef(node)
        node._metadata["flag_type"] = obj
        self.namespace[node.name] = obj

    def visit_EventDef(self, node):
        obj = EventT.from_EventDef(node)
        node._metadata["event_type"] = obj
        self.namespace[node.name] = obj
        self._events.append(obj)

    def visit_FunctionDef(self, node):
        if self.is_interface:
            func_t = ContractFunctionT.from_vyi(node)
            if not func_t.is_external:
                # TODO test me!
                raise StructureException(
                    "Internal functions in `.vyi` files are not allowed!", node
                )
        else:
            func_t = ContractFunctionT.from_FunctionDef(node)

        self._self_t.typ.add_member(func_t.name, func_t)
        node._metadata["func_type"] = func_t
        self._add_exposed_function(func_t, node)

    def visit_Import(self, node):
        # import x.y[name] as y[alias]

        alias = node.alias

        if alias is None:
            alias = node.name

        # don't handle things like `import x.y`
        if "." in alias:
            msg = "import requires an accompanying `as` statement"
            suggested_alias = node.name[node.name.rfind(".") :]
            hint = f"try `import {node.name} as {suggested_alias}`"
            raise StructureException(msg, node, hint=hint)

        self._add_import(node, 0, node.name, alias)

    def visit_ImportFrom(self, node):
        # from m.n[module] import x[name] as y[alias]
        alias = node.alias or node.name

        module = node.module or ""
        if module:
            module += "."

        qualified_module_name = module + node.name
        self._add_import(node, node.level, qualified_module_name, alias)

    def visit_InterfaceDef(self, node):
        interface_t = InterfaceT.from_InterfaceDef(node)
        node._metadata["interface_type"] = interface_t
        self.namespace[node.name] = interface_t

    def visit_StructDef(self, node):
        struct_t = StructT.from_StructDef(node)
        node._metadata["struct_type"] = struct_t
        self.namespace[node.name] = struct_t

    def _add_import(
        self, node: vy_ast.VyperNode, level: int, qualified_module_name: str, alias: str
    ) -> None:
        compiler_input, module_info = self._load_import(node, level, qualified_module_name, alias)
        node._metadata["import_info"] = ImportInfo(
            module_info, alias, qualified_module_name, compiler_input, node
        )
        self.namespace[alias] = module_info

    # load an InterfaceT or ModuleInfo from an import.
    # raises FileNotFoundError
    def _load_import(self, node: vy_ast.VyperNode, level: int, module_str: str, alias: str) -> Any:
        # the directory this (currently being analyzed) module is in
        self_search_path = Path(self.ast.resolved_path).parent

        with self.input_bundle.poke_search_path(self_search_path):
            return self._load_import_helper(node, level, module_str, alias)

    def _load_import_helper(
        self, node: vy_ast.VyperNode, level: int, module_str: str, alias: str
    ) -> tuple[CompilerInput, Any]:
        if _is_builtin(module_str):
            return _load_builtin_import(level, module_str)

        path = _import_to_path(level, module_str)

        # this could conceivably be in the ImportGraph but no need at this point
        if path in self._imported_modules:
            previous_import_stmt = self._imported_modules[path]
            raise DuplicateImport(f"{alias} imported more than once!", previous_import_stmt, node)

        self._imported_modules[path] = node

        err = None

        try:
            path_vy = path.with_suffix(".vy")
            file = self.input_bundle.load_file(path_vy)
            assert isinstance(file, FileInput)  # mypy hint

            module_ast = self._ast_from_file(file)

            with override_global_namespace(Namespace()):
                module_t = _analyze_module_r(
                    module_ast,
                    self.input_bundle,
                    import_graph=self._import_graph,
                    is_interface=False,
                )

                return file, ModuleInfo(module_t, alias)

        except FileNotFoundError as e:
            # escape `e` from the block scope, it can make things
            # easier to debug.
            err = e

        try:
            file = self.input_bundle.load_file(path.with_suffix(".vyi"))
            assert isinstance(file, FileInput)  # mypy hint
            module_ast = self._ast_from_file(file)

            with override_global_namespace(Namespace()):
                _analyze_module_r(
                    module_ast,
                    self.input_bundle,
                    import_graph=self._import_graph,
                    is_interface=True,
                )
                module_t = module_ast._metadata["type"]

                return file, module_t.interface

        except FileNotFoundError:
            pass

        try:
            file = self.input_bundle.load_file(path.with_suffix(".json"))
            assert isinstance(file, ABIInput)  # mypy hint
            return file, InterfaceT.from_json_abi(str(file.path), file.abi)
        except FileNotFoundError:
            pass

        hint = None
        if module_str.startswith("vyper.interfaces"):
            hint = "try renaming `vyper.interfaces` to `ethereum.ercs`"

        # copy search_paths, makes debugging a bit easier
        search_paths = self.input_bundle.search_paths.copy()  # noqa: F841
        raise ModuleNotFound(module_str, hint=hint) from err


def _parse_ast(file: FileInput) -> vy_ast.Module:
    module_path = file.resolved_path  # for error messages
    try:
        # try to get a relative path, to simplify the error message
        cwd = Path(".")
        if module_path.is_absolute():
            cwd = cwd.resolve()
        module_path = module_path.relative_to(cwd)
    except ValueError:
        # we couldn't get a relative path (cf. docs for Path.relative_to),
        # use the resolved path given to us by the InputBundle
        pass

    ret = vy_ast.parse_to_ast(
        file.source_code,
        source_id=file.source_id,
        module_path=module_path.as_posix(),
        resolved_path=file.resolved_path.as_posix(),
    )
    return ret


# convert an import to a path (without suffix)
def _import_to_path(level: int, module_str: str) -> PurePath:
    base_path = ""
    if level > 1:
        base_path = "../" * (level - 1)
    elif level == 1:
        base_path = "./"
    return PurePath(f"{base_path}{module_str.replace('.','/')}/")


# can add more, e.g. "vyper.builtins.interfaces", etc.
BUILTIN_PREFIXES = ["ethereum.ercs"]


# TODO: could move this to analysis/common.py or something
def _is_builtin(module_str):
    return any(module_str.startswith(prefix) for prefix in BUILTIN_PREFIXES)


_builtins_cache: dict[PathLike, tuple[CompilerInput, ModuleT]] = {}


def _load_builtin_import(level: int, module_str: str) -> tuple[CompilerInput, InterfaceT]:
    if not _is_builtin(module_str):  # pragma: nocover
        raise CompilerPanic("unreachable!")

    builtins_path = vyper.builtins.interfaces.__path__[0]
    # hygiene: convert to relpath to avoid leaking user directory info
    # (note Path.relative_to cannot handle absolute to relative path
    # conversion, so we must use the `os` module).
    builtins_path = os.path.relpath(builtins_path)

    search_path = Path(builtins_path).parent.parent.parent
    # generate an input bundle just because it knows how to build paths.
    input_bundle = FilesystemInputBundle([search_path])

    # remap builtins directory --
    # ethereum/ercs => vyper/builtins/interfaces
    remapped_module = module_str
    if remapped_module.startswith("ethereum.ercs"):
        remapped_module = remapped_module.removeprefix("ethereum.ercs")
        remapped_module = vyper.builtins.interfaces.__package__ + remapped_module

    path = _import_to_path(level, remapped_module).with_suffix(".vyi")

    # builtins are globally the same, so we can safely cache them
    # (it is also *correct* to cache them, so that types defined in builtins
    # compare correctly using pointer-equality.)
    if path in _builtins_cache:
        file, module_t = _builtins_cache[path]
        return file, module_t.interface

    try:
        file = input_bundle.load_file(path)
        assert isinstance(file, FileInput)  # mypy hint
    except FileNotFoundError as e:
        hint = None
        components = module_str.split(".")
        # common issue for upgrading codebases from v0.3.x to v0.4.x -
        # hint: rename ERC20 to IERC20
        if components[-1].startswith("ERC"):
            module_prefix = components[-1]
            hint = f"try renaming `{module_prefix}` to `I{module_prefix}`"
        raise ModuleNotFound(module_str, hint=hint) from e

    interface_ast = _parse_ast(file)

    with override_global_namespace(Namespace()):
        module_t = _analyze_module_r(interface_ast, input_bundle, ImportGraph(), is_interface=True)

    _builtins_cache[path] = file, module_t
    return file, module_t.interface



# TODO: rewrite the tests in type-centric way, parametrize array and indices types


def test_negative_ix_access(get_contract, tx_failed):
    # Arrays can't be accessed with negative indices
    code = """
arr: uint256[3]

@external
def foo(i: int128):
    self.arr[i] = 1
    """

    c = get_contract(code)

    with tx_failed():
        c.foo(-1)
    with tx_failed():
        c.foo(-3)
    with tx_failed():
        c.foo(-(2**127) + 1)


def test_negative_ix_access_to_large_arr(get_contract, tx_failed):
    # Arrays can't be accessed with negative indices
    code = """
arr: public(uint256[max_value(uint256)-1])

@external
def set(idx: int256):
    self.arr[idx] = 1
    """

    c = get_contract(code)
    with tx_failed():
        c.set(-(2**255))
    with tx_failed():
        c.set(-(2**255) + 5)
    with tx_failed():
        c.set(-(2**128))
    with tx_failed():
        c.set(-1)


def test_oob_access_to_large_arr(get_contract, tx_failed):
    # Test OOB access to large array
    code = """
arr: public(uint256[max_value(uint256)-1])

@external
def set(idx: int256):
    self.arr[idx] = 3

@external
def set2(idx: uint256):
    self.arr[idx] = 3
    """
    c = get_contract(code)

    with tx_failed():
        c.set2(2**256 - 1)
    with tx_failed():
        c.set2(2**256 - 2)


def test_boundary_access_to_arr(get_contract):
    # Test access to the boundary of the array
    code = """
arr1: public(int256[max_value(int256)])

@external
def set1(idx: int256):
    self.arr1[idx] = 3

    """
    code2 = """
arr2: public(uint256[max_value(uint256)-1])

@external
def set2(idx: uint256):
    self.arr2[idx] = 3
    """
    c1 = get_contract(code)

    c1.set1(2**255 - 2)
    assert c1.arr1(2**255 - 2) == 3
    c1.set1(0)
    assert c1.arr1(0) == 3

    c2 = get_contract(code2)

    c2.set2(2**256 - 3)
    assert c2.arr2(2**256 - 3) == 3


def test_valid_ix_access(get_contract):
    code = """
arr: public(uint256[3])
arr2: public(int256[3])

@external
def foo(i: int128):
    self.arr[i] = 1

@external
def bar(i: uint256):
    self.arr[i] = 2
    """

    c = get_contract(code)
    for i in range(3):
        c.foo(i)
        assert c.arr(i) == 1
        c.bar(i)
        assert c.arr(i) == 2


def test_for_loop_ix_access(get_contract):
    # Arrays can be accessed with for loop iterators of type int
    code = """
arr: public(int256[10])

@external
def foo():
    for i: int256 in range(10):
        self.arr[i] = i
    """

    c = get_contract(code)
    c.foo()
    for i in range(10):
        assert c.arr(i) == i



import pytest

from vyper.evm.opcodes import version_check


@pytest.mark.parametrize("location", ["storage", "transient"])
def test_extract32_extraction(tx_failed, get_contract, location):
    if location == "transient" and not version_check(begin="cancun"):
        pytest.skip(
            "Skipping test as storage_location is 'transient' and EVM version is pre-Cancun"
        )
    if location == "storage":
        decl = "y: Bytes[100]"
    elif location == "transient":
        decl = "y: transient(Bytes[100])"
    else:
        raise Exception("unreachable")
    extract32_code = f"""
{decl}
@external
def extrakt32(inp: Bytes[100], index: uint256) -> bytes32:
    return extract32(inp, index)

@external
def extrakt32_mem(inp: Bytes[100], index: uint256) -> bytes32:
    x: Bytes[100] = inp
    return extract32(x, index)

@external
def extrakt32_storage(index: uint256, inp: Bytes[100]) -> bytes32:
    self.y = inp
    return extract32(self.y, index)
    """

    c = get_contract(extract32_code)
    test_cases = (
        (b"c" * 31, 0),
        (b"c" * 32, 0),
        (b"c" * 33, 0),
        (b"c" * 33, 1),
        (b"c" * 33, 2),
        (b"cow" * 30, 0),
        (b"cow" * 30, 1),
        (b"cow" * 30, 31),
        (b"cow" * 30, 32),
        (b"cow" * 30, 33),
        (b"cow" * 30, 34),
        (b"cow" * 30, 58),
        (b"cow" * 30, 59),
    )

    for S, i in test_cases:
        if 0 <= i <= len(S) - 32:
            expected_result = S[i : i + 32]
            assert c.extrakt32(S, i) == expected_result
            assert c.extrakt32_mem(S, i) == expected_result
            assert c.extrakt32_storage(i, S) == expected_result
        else:
            with tx_failed():
                c.extrakt32(S, i)


def test_extract32_code(tx_failed, get_contract):
    extract32_code = """
@external
def foo(inp: Bytes[32]) -> int128:
    return extract32(inp, 0, output_type=int128)

@external
def bar(inp: Bytes[32]) -> uint256:
    return extract32(inp, 0, output_type=uint256)

@external
def baz(inp: Bytes[32]) -> bytes32:
    return extract32(inp, 0, output_type=bytes32)

@external
def fop(inp: Bytes[32]) -> bytes32:
    return extract32(inp, 0)

@external
def foq(inp: Bytes[32]) -> address:
    return extract32(inp, 0, output_type=address)
    """

    c = get_contract(extract32_code)
    assert c.foo(b"\x00" * 30 + b"\x01\x01") == 257
    assert c.bar(b"\x00" * 30 + b"\x01\x01") == 257

    with tx_failed():
        c.foo(b"\x80" + b"\x00" * 30)

    assert c.bar(b"\x80" + b"\x00" * 31) == 2**255

    assert c.baz(b"crow" * 8) == b"crow" * 8
    assert c.fop(b"crow" * 8) == b"crow" * 8
    assert c.foq(b"\x00" * 12 + b"3" * 20) == "0x" + "3" * 40

    with tx_failed():
        c.foq(b"crow" * 8)



import pytest
import rlp
from eth.codecs import abi
from hexbytes import HexBytes

import vyper.ir.compile_ir as compile_ir
from tests.utils import ZERO_ADDRESS
from vyper.codegen.ir_node import IRnode
from vyper.compiler.settings import OptimizationLevel
from vyper.utils import EIP_170_LIMIT, ERC5202_PREFIX, checksum_encode, keccak256


# initcode used by create_minimal_proxy_to
def eip1167_initcode(_addr):
    addr = HexBytes(_addr)
    pre = HexBytes("0x602D3D8160093D39F3363d3d373d3d3d363d73")
    post = HexBytes("0x5af43d82803e903d91602b57fd5bf3")
    return HexBytes(pre + (addr + HexBytes(0) * (20 - len(addr))) + post)


# initcode used by CreateCopyOf
def vyper_initcode(runtime_bytecode):
    bytecode_len_hex = hex(len(runtime_bytecode))[2:].rjust(6, "0")
    return HexBytes("0x62" + bytecode_len_hex + "3d81600b3d39f3") + runtime_bytecode


def test_create_minimal_proxy_to_create(get_contract):
    code = """
main: address

@external
def test() -> address:
    self.main = create_minimal_proxy_to(self)
    return self.main
    """

    c = get_contract(code)

    address_bits = int(c.address, 16)
    nonce = 1
    rlp_encoded = rlp.encode([address_bits, nonce])
    expected_create_address = keccak256(rlp_encoded)[12:].rjust(20, b"\x00")
    assert c.test() == checksum_encode("0x" + expected_create_address.hex())


def test_create_minimal_proxy_to_call(get_contract):
    code = """
interface SubContract:
    def hello() -> Bytes[100]: view

other: public(address)

@external
def test() -> address:
    self.other = create_minimal_proxy_to(self)
    return self.other

@external
def hello() -> Bytes[100]:
    return b"hello world!"

@external
def test2() -> Bytes[100]:
    return staticcall SubContract(self.other).hello()
    """

    c = get_contract(code)

    assert c.hello() == b"hello world!"
    c.test()
    assert c.test2() == b"hello world!"


def test_minimal_proxy_exception(env, get_contract, tx_failed):
    code = """
interface SubContract:
    def hello(a: uint256) -> Bytes[100]: view

other: public(address)

@external
def test() -> address:
    self.other = create_minimal_proxy_to(self)
    return self.other

@external
def hello(a: uint256) -> Bytes[100]:
    assert a > 0, "invaliddddd"
    return b"hello world!"

@external
def test2(a: uint256) -> Bytes[100]:
    return staticcall SubContract(self.other).hello(a)
    """

    c = get_contract(code)

    assert c.hello(1) == b"hello world!"
    c.test()
    assert c.test2(1) == b"hello world!"

    with tx_failed(exc_text="invaliddddd"):
        c.test2(0)

    gas_sent = 30000
    with tx_failed(exc_text="invaliddddd"):
        c.test2(0, gas=gas_sent)

    # check we issued `revert`, which does not consume all gas
    assert env.last_result.gas_used < gas_sent


@pytest.mark.parametrize("revert_on_failure", [True, False, None])
def test_create_minimal_proxy_to_create2(
    get_contract, create2_address_of, keccak, tx_failed, revert_on_failure
):
    revert_arg = "" if revert_on_failure is None else f", revert_on_failure={revert_on_failure}"
    code = f"""
main: address

@external
def test(_salt: bytes32) -> address:
    self.main = create_minimal_proxy_to(self, salt=_salt{revert_arg})
    return self.main
    """

    c = get_contract(code)

    salt = keccak(b"vyper")
    result = c.test(salt)
    assert HexBytes(result) == create2_address_of(c.address, salt, eip1167_initcode(c.address))

    # revert on collision
    if revert_on_failure is False:
        assert c.test(salt) == ZERO_ADDRESS
    else:
        with tx_failed():
            c.test(salt)


# test blueprints with various prefixes - 0xfe would block calls to the blueprint
# contract, and 0xfe7100 is ERC5202 magic
@pytest.mark.parametrize("blueprint_prefix", [b"", b"\xfe", ERC5202_PREFIX])
@pytest.mark.parametrize("revert_on_failure", [True, False, None])
def test_create_from_blueprint(
    get_contract,
    deploy_blueprint_for,
    env,
    keccak,
    create2_address_of,
    tx_failed,
    blueprint_prefix,
    revert_on_failure,
):
    revert_arg = "" if revert_on_failure is None else f", revert_on_failure={revert_on_failure}"
    code = """
@external
def foo() -> uint256:
    return 123
    """

    prefix_len = len(blueprint_prefix)
    deployer_code = f"""
created_address: public(address)

@external
def test(target: address):
    self.created_address = create_from_blueprint(target, code_offset={prefix_len}{revert_arg})

@external
def test2(target: address, salt: bytes32):
    self.created_address = create_from_blueprint(
        target, code_offset={prefix_len}, salt=salt{revert_arg}
    )
    """

    # deploy a foo, so we can compare its bytecode with factory deployed version
    foo_contract = get_contract(code)
    expected_runtime_code = env.get_code(foo_contract.address)

    f, FooContract = deploy_blueprint_for(code, initcode_prefix=blueprint_prefix)

    d = get_contract(deployer_code)

    d.test(f.address)

    test = FooContract(d.created_address())
    assert env.get_code(test.address) == expected_runtime_code
    assert test.foo() == 123

    # extcodesize check
    zero_address = "0x" + "00" * 20
    with tx_failed():
        d.test(zero_address)

    # now same thing but with create2
    salt = keccak(b"vyper")
    d.test2(f.address, salt)

    test = FooContract(d.created_address())
    assert env.get_code(test.address) == expected_runtime_code
    assert test.foo() == 123

    # check if the create2 address matches our offchain calculation
    initcode = env.get_code(f.address)
    initcode = initcode[len(blueprint_prefix) :]  # strip the prefix
    assert HexBytes(test.address) == create2_address_of(d.address, salt, initcode)

    # can't collide addresses
    if revert_on_failure is False:
        assert not d.test2(f.address, salt)
    else:
        with tx_failed():
            d.test2(f.address, salt)


# test blueprints with 0xfe7100 prefix, which is the EIP 5202 standard.
# code offset by default should be 3 here.
def test_create_from_blueprint_default_offset(
    get_contract, deploy_blueprint_for, env, keccak, create2_address_of, tx_failed
):
    code = """
@external
def foo() -> uint256:
    return 123
    """

    deployer_code = """
created_address: public(address)

@external
def test(target: address):
    self.created_address = create_from_blueprint(target)

@external
def test2(target: address, salt: bytes32):
    self.created_address = create_from_blueprint(target, salt=salt)
    """

    # deploy a foo so we can compare its bytecode with factory deployed version
    foo_contract = get_contract(code)
    expected_runtime_code = env.get_code(foo_contract.address)

    f, FooContract = deploy_blueprint_for(code)

    d = get_contract(deployer_code)

    d.test(f.address)

    test = FooContract(d.created_address())
    assert env.get_code(test.address) == expected_runtime_code
    assert test.foo() == 123

    # extcodesize check
    zero_address = "0x" + "00" * 20
    with tx_failed():
        d.test(zero_address)

    # now same thing but with create2
    salt = keccak(b"vyper")
    d.test2(f.address, salt)

    test = FooContract(d.created_address())
    assert env.get_code(test.address) == expected_runtime_code
    assert test.foo() == 123

    # check if the create2 address matches our offchain calculation
    initcode = env.get_code(f.address)
    initcode = initcode[len(ERC5202_PREFIX) :]  # strip the prefix
    assert HexBytes(test.address) == create2_address_of(d.address, salt, initcode)

    # can't collide addresses
    with tx_failed():
        d.test2(f.address, salt)


def test_create_from_blueprint_bad_code_offset(
    get_contract, get_contract_from_ir, deploy_blueprint_for, env, tx_failed
):
    deployer_code = """
BLUEPRINT: immutable(address)

@deploy
def __init__(blueprint_address: address):
    BLUEPRINT = blueprint_address

@external
def test(code_ofst: uint256) -> address:
    return create_from_blueprint(BLUEPRINT, code_offset=code_ofst)
    """

    initcode_len = 100

    # deploy a blueprint contract whose contained initcode contains only
    # zeroes (so no matter which offset, create_from_blueprint will
    # return empty code)
    ir = IRnode.from_list(["deploy", 0, ["seq"] + ["stop"] * initcode_len, 0])
    bytecode, _ = compile_ir.assembly_to_evm(
        compile_ir.compile_to_assembly(ir, optimize=OptimizationLevel.NONE)
    )
    # manually deploy the bytecode
    c = env.deploy(abi=[], bytecode=bytecode)
    blueprint_address = c.address

    d = get_contract(deployer_code, blueprint_address)

    # deploy with code_ofst=0 fine
    d.test(0)

    # deploy with code_ofst=len(blueprint) - 1 fine
    d.test(initcode_len - 1)

    # code_offset=len(blueprint) NOT fine! would EXTCODECOPY empty initcode
    with tx_failed():
        d.test(initcode_len)

    # code_offset=EIP_170_LIMIT definitely not fine!
    with tx_failed():
        d.test(EIP_170_LIMIT)


# test create_from_blueprint with args
def test_create_from_blueprint_args(
    get_contract, deploy_blueprint_for, env, keccak, create2_address_of, tx_failed
):
    code = """
struct Bar:
    x: String[32]

FOO: immutable(String[128])
BAR: immutable(Bar)

@deploy
def __init__(foo: String[128], bar: Bar):
    FOO = foo
    BAR = bar

@external
def foo() -> String[128]:
    return FOO

@external
def bar() -> Bar:
    return BAR
    """

    deployer_code = """
struct Bar:
    x: String[32]

created_address: public(address)

@external
def test(target: address, arg1: String[128], arg2: Bar):
    self.created_address = create_from_blueprint(target, arg1, arg2)

@external
def test2(target: address, arg1: String[128], arg2: Bar, salt: bytes32):
    self.created_address = create_from_blueprint(target, arg1, arg2, salt=salt)

@external
def test3(target: address, argdata: Bytes[1024]):
    self.created_address = create_from_blueprint(target, argdata, raw_args=True)

@external
def test4(target: address, argdata: Bytes[1024], salt: bytes32):
    self.created_address = create_from_blueprint(target, argdata, salt=salt, raw_args=True)

@external
def should_fail(target: address, arg1: String[129], arg2: Bar):
    self.created_address = create_from_blueprint(target, arg1, arg2)
    """
    FOO = "hello!"
    BAR = ("world!",)

    # deploy a foo so we can compare its bytecode with factory deployed version
    foo_contract = get_contract(code, FOO, BAR)
    expected_runtime_code = env.get_code(foo_contract.address)

    f, FooContract = deploy_blueprint_for(code)

    d = get_contract(deployer_code)

    initcode = env.get_code(f.address)[3:]

    d.test(f.address, FOO, BAR)

    test = FooContract(d.created_address())
    assert env.get_code(test.address) == expected_runtime_code
    assert test.foo() == FOO
    assert test.bar() == BAR

    # extcodesize check
    with tx_failed():
        d.test("0x" + "00" * 20, FOO, BAR)

    # now same thing but with create2
    salt = keccak(b"vyper")
    d.test2(f.address, FOO, BAR, salt)

    test = FooContract(d.created_address())
    assert env.get_code(test.address) == expected_runtime_code
    assert test.foo() == FOO
    assert test.bar() == BAR

    encoded_args = abi.encode("(string,(string))", (FOO, BAR))
    assert HexBytes(test.address) == create2_address_of(d.address, salt, initcode + encoded_args)

    d.test3(f.address, encoded_args)
    test = FooContract(d.created_address())
    assert env.get_code(test.address) == expected_runtime_code
    assert test.foo() == FOO
    assert test.bar() == BAR

    d.test4(f.address, encoded_args, keccak(b"test4"))
    test = FooContract(d.created_address())
    assert env.get_code(test.address) == expected_runtime_code
    assert test.foo() == FOO
    assert test.bar() == BAR

    # can't collide addresses
    with tx_failed():
        d.test2(f.address, FOO, BAR, salt)
    # ditto - with raw_args
    with tx_failed():
        d.test4(f.address, encoded_args, salt)

    # but creating a contract with different args is ok
    FOO = "bar"
    d.test2(f.address, FOO, BAR, salt)
    # just for kicks
    assert FooContract(d.created_address()).foo() == FOO
    assert FooContract(d.created_address()).bar() == BAR

    # Foo constructor should fail
    FOO = "01" * 129
    BAR = ("",)
    sig = keccak("should_fail(address,string,(string))".encode()).hex()[:10]
    encoded = abi.encode("(address,string,(string))", (f.address, FOO, BAR)).hex()
    with tx_failed():
        env.message_call(d.address, env.deployer, f"{sig}{encoded}")


@pytest.mark.parametrize("revert_on_failure", [True, False, None])
def test_create_copy_of(
    get_contract, env, keccak, create2_address_of, tx_failed, revert_on_failure
):
    revert_arg = "" if revert_on_failure is None else f", revert_on_failure={revert_on_failure}"
    code = f"""
created_address: public(address)
@internal
def _create_copy_of(target: address):
    self.created_address = create_copy_of(target{revert_arg})

@internal
def _create_copy_of2(target: address, salt: bytes32):
    self.created_address = create_copy_of(target, salt=salt{revert_arg})

@external
def test(target: address) -> address:
    x: uint256 = 0
    self._create_copy_of(target)
    assert x == 0  # check memory not clobbered
    return self.created_address

@external
def test2(target: address, salt: bytes32) -> address:
    x: uint256 = 0
    self._create_copy_of2(target, salt)
    assert x == 0  # check memory not clobbered
    return self.created_address
    """

    c = get_contract(code)
    bytecode = env.get_code(c.address)

    c.test(c.address)
    test1 = c.created_address()
    assert env.get_code(test1) == bytecode

    # extcodesize check
    with tx_failed():
        c.test("0x" + "00" * 20)

    salt = keccak(b"vyper")
    c.test2(c.address, salt)
    test2 = c.created_address()
    assert env.get_code(test2) == bytecode

    assert HexBytes(test2) == create2_address_of(c.address, salt, vyper_initcode(bytecode))

    # can't create2 where contract already exists
    if revert_on_failure is False:
        assert c.test2(c.address, salt) == ZERO_ADDRESS
    else:
        with tx_failed():
            c.test2(c.address, salt)


# XXX: these various tests to check the msize allocator for
# create_copy_of and create_from_blueprint depend on calling convention
# and variables writing to memory. think of ways to make more robust to
# changes in calling convention and memory layout
@pytest.mark.parametrize("blueprint_prefix", [b"", b"\xfe", b"\xfe\71\x00"])
def test_create_from_blueprint_complex_value(
    get_contract, deploy_blueprint_for, env, blueprint_prefix
):
    # check msize allocator does not get trampled by value= kwarg
    code = """
var: uint256

@deploy
@payable
def __init__(x: uint256):
    self.var = x

@external
def foo()-> uint256:
    return self.var
    """

    prefix_len = len(blueprint_prefix)

    some_constant = b"\00" * 31 + b"\x0c"

    deployer_code = f"""
created_address: public(address)
x: constant(Bytes[32]) = {some_constant}

@internal
def foo() -> uint256:
    g:uint256 = 42
    return 3

@external
@payable
def test(target: address):
    self.created_address = create_from_blueprint(
        target,
        x,
        code_offset={prefix_len},
        value=self.foo(),
        raw_args=True
    )
    """

    foo_contract = get_contract(code, 12)
    expected_runtime_code = env.get_code(foo_contract.address)

    f, FooContract = deploy_blueprint_for(code, initcode_prefix=blueprint_prefix)

    d = get_contract(deployer_code)

    env.set_balance(env.deployer, 3)
    d.test(f.address, value=3)

    test = FooContract(d.created_address())
    assert env.get_code(test.address) == expected_runtime_code
    assert test.foo() == 12


@pytest.mark.parametrize("blueprint_prefix", [b"", b"\xfe", b"\xfe\71\x00"])
def test_create_from_blueprint_complex_salt_raw_args(
    get_contract, deploy_blueprint_for, env, blueprint_prefix
):
    # test msize allocator does not get trampled by salt= kwarg
    code = """
var: uint256

@deploy
@payable
def __init__(x: uint256):
    self.var = x

@external
def foo()-> uint256:
    return self.var
    """

    some_constant = b"\00" * 31 + b"\x0c"
    prefix_len = len(blueprint_prefix)

    deployer_code = f"""
created_address: public(address)

x: constant(Bytes[32]) = {some_constant}
salt: constant(bytes32) = keccak256("kebab")

@internal
def foo() -> bytes32:
    g:uint256 = 42
    return salt

@external
@payable
def test(target: address):
    self.created_address = create_from_blueprint(
        target,
        x,
        code_offset={prefix_len},
        salt=self.foo(),
        raw_args= True
    )
    """

    foo_contract = get_contract(code, 12)
    expected_runtime_code = env.get_code(foo_contract.address)

    f, FooContract = deploy_blueprint_for(code, initcode_prefix=blueprint_prefix)

    d = get_contract(deployer_code)

    d.test(f.address)

    test = FooContract(d.created_address())
    assert env.get_code(test.address) == expected_runtime_code
    assert test.foo() == 12


@pytest.mark.parametrize("blueprint_prefix", [b"", b"\xfe", b"\xfe\71\x00"])
def test_create_from_blueprint_complex_salt_no_constructor_args(
    get_contract, deploy_blueprint_for, env, blueprint_prefix
):
    # test msize allocator does not get trampled by salt= kwarg
    code = """
var: uint256

@deploy
@payable
def __init__():
    self.var = 12

@external
def foo()-> uint256:
    return self.var
    """

    prefix_len = len(blueprint_prefix)
    deployer_code = f"""
created_address: public(address)

salt: constant(bytes32) = keccak256("kebab")

@external
@payable
def test(target: address):
    self.created_address = create_from_blueprint(
        target,
        code_offset={prefix_len},
        salt=keccak256(_abi_encode(target))
    )
    """

    foo_contract = get_contract(code)
    expected_runtime_code = env.get_code(foo_contract.address)

    f, FooContract = deploy_blueprint_for(code, initcode_prefix=blueprint_prefix)

    d = get_contract(deployer_code)

    d.test(f.address)

    test = FooContract(d.created_address())
    assert env.get_code(test.address) == expected_runtime_code
    assert test.foo() == 12


def test_create_copy_of_complex_kwargs(get_contract, env):
    # test msize allocator does not get trampled by salt= kwarg
    complex_salt = """
created_address: public(address)

@external
def test(target: address) -> address:
    self.created_address = create_copy_of(
        target,
        salt=keccak256(_abi_encode(target))
    )
    return self.created_address

    """

    c = get_contract(complex_salt)
    bytecode = env.get_code(c.address)
    assert bytecode  # Sanity check
    c.test(c.address)
    test1 = c.address
    assert env.get_code(test1) == bytecode

    # test msize allocator does not get trampled by value= kwarg
    complex_value = """
created_address: public(address)

@external
@payable
def test(target: address) -> address:
    value: uint256 = 2
    self.created_address = create_copy_of(target, value = [2,2,2][value])
    return self.created_address

    """

    c = get_contract(complex_value)
    bytecode = env.get_code(c.address)
    env.set_balance(env.deployer, 2)

    c.test(c.address, value=2)
    test1 = c.address
    assert env.get_code(test1) == bytecode



from typing import List

from vyper.exceptions import CompilerPanic, MemoryAllocationException
from vyper.utils import MemoryPositions


class FreeMemory:
    __slots__ = ("position", "size")

    def __init__(self, position: int, size: int) -> None:
        self.position = position
        self.size = size

    def __repr__(self):
        return f"(FreeMemory: pos={self.position}, size={self.size})"

    def partially_allocate(self, size: int) -> int:
        """
        Reduce the size of the free memory by allocating from the initial offset.

        Arguments
        ---------
        size : int
            Number of bytes to allocate

        Returns
        -------
        int
            Position of the newly allocated memory
        """
        if size >= self.size:
            raise CompilerPanic("Attempted to allocate more memory than available")
        position = self.position
        self.position += size
        self.size -= size
        return position


class MemoryAllocator:
    """
    Low-level memory allocator. Used to allocate and de-allocate memory slots.

    This object should not be accessed directly. Memory allocation happens via
    declaring variables within `Context`.
    """

    next_mem: int

    _ALLOCATION_LIMIT: int = 2**64

    def __init__(self, start_position: int = MemoryPositions.RESERVED_MEMORY):
        """
        Initializer.

        Arguments
        ---------
        start_position : int, optional
            The initial offset to use as the free memory pointer. Offsets
            prior to this value are considered permanently allocated.
        """
        self.next_mem = start_position
        self.size_of_mem = start_position
        self.deallocated_mem: List[FreeMemory] = []

    # Get the next unused memory location
    def get_next_memory_position(self) -> int:
        return self.next_mem

    def allocate_memory(self, size: int) -> int:
        """
        Allocate `size` bytes in memory.

        *** No guarantees are made that allocated memory is clean! ***

        If no memory was previously de-allocated, memory is expanded
        and the free memory pointer is increased.

        If sufficient space is available within de-allocated memory, the lowest
        available offset is returned and that memory is now marked as allocated.

        Arguments
        ---------
        size : int
            The number of bytes to allocate. Must be divisible by 32.

        Returns
        -------
        int
            Start offset of the newly allocated memory.
        """
        if size % 32 != 0:
            raise CompilerPanic(f"tried to allocate {size} bytes, only multiples of 32 supported.")

        # check for deallocated memory prior to expanding
        for i, free_memory in enumerate(self.deallocated_mem):
            if free_memory.size == size:
                del self.deallocated_mem[i]
                return free_memory.position
            if free_memory.size > size:
                return free_memory.partially_allocate(size)

        # if no deallocated slots are available, expand memory
        return self._expand_memory(size)

    def _expand_memory(self, size: int) -> int:
        """
        Allocate `size` bytes in memory, starting from the free memory pointer.
        """
        if size % 32 != 0:
            raise CompilerPanic("Memory misaligment, only multiples of 32 supported.")

        before_value = self.next_mem
        self.next_mem += size
        self.size_of_mem = max(self.size_of_mem, self.next_mem)

        if self.size_of_mem >= self._ALLOCATION_LIMIT:
            # this should not be caught
            raise MemoryAllocationException(
                f"Tried to allocate {self.size_of_mem} bytes! "
                f"(limit is {self._ALLOCATION_LIMIT} (2**64) bytes)"
            )

        return before_value

    def deallocate_memory(self, pos: int, size: int) -> None:
        """
        De-allocate memory.

        Arguments
        ---------
        pos : int
            The initial memory position to de-allocate.
        size : int
            The number of bytes to de-allocate. Must be divisible by 32.
        """
        if size % 32 != 0:
            raise CompilerPanic("Memory misaligment, only multiples of 32 supported.")

        self.deallocated_mem.append(FreeMemory(position=pos, size=size))
        self.deallocated_mem.sort(key=lambda k: k.position)

        if not self.deallocated_mem:
            return

        # iterate over deallocated memory and merge slots where possible
        i = 1
        active = self.deallocated_mem[0]
        while len(self.deallocated_mem) > i:
            next_slot = self.deallocated_mem[i]
            if next_slot.position == active.position + active.size:
                active.size += next_slot.size
                self.deallocated_mem.remove(next_slot)
            else:
                active = next_slot
                i += 1

        # if the highest free memory slot ends at the edge of the
        # allocated memory, reduce the free memory pointer
        last = self.deallocated_mem[-1]
        if last.position + last.size == self.next_mem:
            self.next_mem = last.position
            del self.deallocated_mem[-1]



from collections import Counter
from typing import Any

from vyper.exceptions import CompilerPanic, StackTooDeep
from vyper.ir.compile_ir import (
    PUSH,
    DataHeader,
    Instruction,
    RuntimeHeader,
    mksymbol,
    optimize_assembly,
)
from vyper.utils import MemoryPositions, OrderedSet
from vyper.venom.analysis.analysis import IRAnalysesCache
from vyper.venom.analysis.dup_requirements import DupRequirementsAnalysis
from vyper.venom.analysis.liveness import LivenessAnalysis
from vyper.venom.basicblock import (
    IRBasicBlock,
    IRInstruction,
    IRLabel,
    IRLiteral,
    IROperand,
    IRVariable,
)
from vyper.venom.context import IRContext
from vyper.venom.passes.normalization import NormalizationPass
from vyper.venom.stack_model import StackModel

# instructions which map one-to-one from venom to EVM
_ONE_TO_ONE_INSTRUCTIONS = frozenset(
    [
        "revert",
        "coinbase",
        "calldatasize",
        "calldatacopy",
        "mcopy",
        "calldataload",
        "gas",
        "gasprice",
        "gaslimit",
        "chainid",
        "address",
        "origin",
        "number",
        "extcodesize",
        "extcodehash",
        "extcodecopy",
        "returndatasize",
        "returndatacopy",
        "callvalue",
        "selfbalance",
        "sload",
        "sstore",
        "mload",
        "mstore",
        "tload",
        "tstore",
        "timestamp",
        "caller",
        "blockhash",
        "selfdestruct",
        "signextend",
        "stop",
        "shr",
        "shl",
        "sar",
        "and",
        "xor",
        "or",
        "add",
        "sub",
        "mul",
        "div",
        "smul",
        "sdiv",
        "mod",
        "smod",
        "exp",
        "addmod",
        "mulmod",
        "eq",
        "iszero",
        "not",
        "lg",
        "lt",
        "slt",
        "sgt",
        "create",
        "create2",
        "msize",
        "balance",
        "call",
        "staticcall",
        "delegatecall",
        "codesize",
        "basefee",
        "blobhash",
        "blobbasefee",
        "prevrandao",
        "difficulty",
        "invalid",
    ]
)

_REVERT_POSTAMBLE = ["_sym___revert", "JUMPDEST", *PUSH(0), "DUP1", "REVERT"]


def apply_line_numbers(inst: IRInstruction, asm) -> list[str]:
    ret = []
    for op in asm:
        if isinstance(op, str) and not isinstance(op, Instruction):
            ret.append(Instruction(op, inst.ast_source, inst.error_msg))
        else:
            ret.append(op)
    return ret  # type: ignore


# TODO: "assembly" gets into the recursion due to how the original
# IR was structured recursively in regards with the deploy instruction.
# There, recursing into the deploy instruction was by design, and
# made it easier to make the assembly generated "recursive" (i.e.
# instructions being lists of instructions). We don't have this restriction
# anymore, so we can probably refactor this to be iterative in coordination
# with the assembler. My suggestion is to let this be for now, and we can
# refactor it later when we are finished phasing out the old IR.
class VenomCompiler:
    ctxs: list[IRContext]
    label_counter = 0
    visited_instructions: OrderedSet  # {IRInstruction}
    visited_basicblocks: OrderedSet  # {IRBasicBlock}
    liveness_analysis: LivenessAnalysis

    def __init__(self, ctxs: list[IRContext]):
        self.ctxs = ctxs
        self.label_counter = 0
        self.visited_instructions = OrderedSet()
        self.visited_basicblocks = OrderedSet()

    def generate_evm(self, no_optimize: bool = False) -> list[str]:
        self.visited_instructions = OrderedSet()
        self.visited_basicblocks = OrderedSet()
        self.label_counter = 0

        asm: list[Any] = []
        top_asm = asm

        for ctx in self.ctxs:
            for fn in ctx.functions.values():
                ac = IRAnalysesCache(fn)

                NormalizationPass(ac, fn).run_pass()
                self.liveness_analysis = ac.request_analysis(LivenessAnalysis)
                ac.request_analysis(DupRequirementsAnalysis)

                assert fn.normalized, "Non-normalized CFG!"

                self._generate_evm_for_basicblock_r(asm, fn.entry, StackModel())

            # TODO make this property on IRFunction
            asm.extend(["_sym__ctor_exit", "JUMPDEST"])
            if ctx.immutables_len is not None and ctx.ctor_mem_size is not None:
                asm.extend(
                    ["_sym_subcode_size", "_sym_runtime_begin", "_mem_deploy_start", "CODECOPY"]
                )
                asm.extend(["_OFST", "_sym_subcode_size", ctx.immutables_len])  # stack: len
                asm.extend(["_mem_deploy_start"])  # stack: len mem_ofst
                asm.extend(["RETURN"])
                asm.extend(_REVERT_POSTAMBLE)
                runtime_asm = [
                    RuntimeHeader("_sym_runtime_begin", ctx.ctor_mem_size, ctx.immutables_len)
                ]
                asm.append(runtime_asm)
                asm = runtime_asm
            else:
                asm.extend(_REVERT_POSTAMBLE)

            # Append data segment
            data_segments: dict = dict()
            for inst in ctx.data_segment:
                if inst.opcode == "dbname":
                    label = inst.operands[0].value
                    data_segments[label] = [DataHeader(f"_sym_{label}")]
                elif inst.opcode == "db":
                    data = inst.operands[0]
                    if isinstance(data, IRLabel):
                        data_segments[label].append(f"_sym_{data.value}")
                    else:
                        data_segments[label].append(data)

            asm.extend(list(data_segments.values()))

        if no_optimize is False:
            optimize_assembly(top_asm)

        return top_asm

    def _stack_reorder(
        self, assembly: list, stack: StackModel, stack_ops: list[IRVariable]
    ) -> None:
        stack_ops_count = len(stack_ops)

        counts = Counter(stack_ops)

        for i in range(stack_ops_count):
            op = stack_ops[i]
            final_stack_depth = -(stack_ops_count - i - 1)
            depth = stack.get_depth(op, counts[op])  # type: ignore
            counts[op] -= 1

            if depth == StackModel.NOT_IN_STACK:
                raise CompilerPanic(f"Variable {op} not in stack")

            if depth == final_stack_depth:
                continue

            if op == stack.peek(final_stack_depth):
                continue

            self.swap(assembly, stack, depth)
            self.swap(assembly, stack, final_stack_depth)

    def _emit_input_operands(
        self, assembly: list, inst: IRInstruction, ops: list[IROperand], stack: StackModel
    ) -> None:
        # PRE: we already have all the items on the stack that have
        # been scheduled to be killed. now it's just a matter of emitting
        # SWAPs, DUPs and PUSHes until we match the `ops` argument

        # dumb heuristic: if the top of stack is not wanted here, swap
        # it with something that is wanted
        if ops and stack.height > 0 and stack.peek(0) not in ops:
            for op in ops:
                if isinstance(op, IRVariable) and op not in inst.dup_requirements:
                    self.swap_op(assembly, stack, op)
                    break

        emitted_ops = OrderedSet[IROperand]()
        for op in ops:
            if isinstance(op, IRLabel):
                # invoke emits the actual instruction itself so we don't need to emit it here
                # but we need to add it to the stack map
                if inst.opcode != "invoke":
                    assembly.append(f"_sym_{op.value}")
                stack.push(op)
                continue

            if isinstance(op, IRLiteral):
                if op.value < -(2**255):
                    raise Exception(f"Value too low: {op.value}")
                elif op.value >= 2**256:
                    raise Exception(f"Value too high: {op.value}")
                assembly.extend(PUSH(op.value % 2**256))
                stack.push(op)
                continue

            if op in inst.dup_requirements and op not in emitted_ops:
                self.dup_op(assembly, stack, op)

            if op in emitted_ops:
                self.dup_op(assembly, stack, op)

            emitted_ops.add(op)

    def _generate_evm_for_basicblock_r(
        self, asm: list, basicblock: IRBasicBlock, stack: StackModel
    ) -> None:
        if basicblock in self.visited_basicblocks:
            return
        self.visited_basicblocks.add(basicblock)

        # assembly entry point into the block
        asm.append(f"_sym_{basicblock.label}")
        asm.append("JUMPDEST")

        self.clean_stack_from_cfg_in(asm, basicblock, stack)

        param_insts = [inst for inst in basicblock.instructions if inst.opcode == "param"]
        main_insts = [inst for inst in basicblock.instructions if inst.opcode != "param"]

        for inst in param_insts:
            asm.extend(self._generate_evm_for_instruction(inst, stack))

        self._clean_unused_params(asm, basicblock, stack)

        for i, inst in enumerate(main_insts):
            next_liveness = main_insts[i + 1].liveness if i + 1 < len(main_insts) else OrderedSet()

            asm.extend(self._generate_evm_for_instruction(inst, stack, next_liveness))

        for bb in basicblock.reachable:
            self._generate_evm_for_basicblock_r(asm, bb, stack.copy())

    def _clean_unused_params(self, asm: list, bb: IRBasicBlock, stack: StackModel) -> None:
        for i, inst in enumerate(bb.instructions):
            if inst.opcode != "param":
                break
            if inst.volatile and i + 1 < len(bb.instructions):
                liveness = bb.instructions[i + 1].liveness
                if inst.output is not None and inst.output not in liveness:
                    depth = stack.get_depth(inst.output)
                    if depth != 0:
                        self.swap(asm, stack, depth)
                    self.pop(asm, stack)

    # pop values from stack at entry to bb
    # note this produces the same result(!) no matter which basic block
    # we enter from in the CFG.
    def clean_stack_from_cfg_in(
        self, asm: list, basicblock: IRBasicBlock, stack: StackModel
    ) -> None:
        if len(basicblock.cfg_in) == 0:
            return

        to_pop = OrderedSet[IRVariable]()
        for in_bb in basicblock.cfg_in:
            # inputs is the input variables we need from in_bb
            inputs = self.liveness_analysis.input_vars_from(in_bb, basicblock)

            # layout is the output stack layout for in_bb (which works
            # for all possible cfg_outs from the in_bb).
            layout = in_bb.out_vars

            # pop all the stack items which in_bb produced which we don't need.
            to_pop |= layout.difference(inputs)

        for var in to_pop:
            depth = stack.get_depth(var)
            # don't pop phantom phi inputs
            if depth is StackModel.NOT_IN_STACK:
                continue

            if depth != 0:
                self.swap(asm, stack, depth)
            self.pop(asm, stack)

    def _generate_evm_for_instruction(
        self, inst: IRInstruction, stack: StackModel, next_liveness: OrderedSet = None
    ) -> list[str]:
        assembly: list[str | int] = []
        next_liveness = next_liveness or OrderedSet()
        opcode = inst.opcode

        #
        # generate EVM for op
        #

        # Step 1: Apply instruction special stack manipulations

        if opcode in ["jmp", "djmp", "jnz", "invoke"]:
            operands = list(inst.get_non_label_operands())
        elif opcode == "alloca":
            offset, _size = inst.operands
            operands = [offset]

        # iload and istore are special cases because they can take a literal
        # that is handled specialy with the _OFST macro. Look below, after the
        # stack reordering.
        elif opcode == "iload":
            addr = inst.operands[0]
            if isinstance(addr, IRLiteral):
                operands = []
            else:
                operands = inst.operands
        elif opcode == "istore":
            addr = inst.operands[1]
            if isinstance(addr, IRLiteral):
                operands = inst.operands[:1]
            else:
                operands = inst.operands
        elif opcode == "log":
            log_topic_count = inst.operands[0].value
            assert log_topic_count in [0, 1, 2, 3, 4], "Invalid topic count"
            operands = inst.operands[1:]
        else:
            operands = inst.operands

        if opcode == "phi":
            ret = inst.get_outputs()[0]
            phis = list(inst.get_inputs())
            depth = stack.get_phi_depth(phis)
            # collapse the arguments to the phi node in the stack.
            # example, for `%56 = %label1 %13 %label2 %14`, we will
            # find an instance of %13 *or* %14 in the stack and replace it with %56.
            to_be_replaced = stack.peek(depth)
            if to_be_replaced in inst.dup_requirements:
                # %13/%14 is still live(!), so we make a copy of it
                self.dup(assembly, stack, depth)
                stack.poke(0, ret)
            else:
                stack.poke(depth, ret)
            return apply_line_numbers(inst, assembly)

        # Step 2: Emit instruction's input operands
        self._emit_input_operands(assembly, inst, operands, stack)

        # Step 3: Reorder stack
        if opcode in ["jnz", "djmp", "jmp"]:
            # prepare stack for jump into another basic block
            assert inst.parent and isinstance(inst.parent.cfg_out, OrderedSet)
            b = next(iter(inst.parent.cfg_out))
            target_stack = self.liveness_analysis.input_vars_from(inst.parent, b)
            # TODO optimize stack reordering at entry and exit from basic blocks
            # NOTE: stack in general can contain multiple copies of the same variable,
            # however we are safe in the case of jmp/djmp/jnz as it's not going to
            # have multiples.
            target_stack_list = list(target_stack)
            self._stack_reorder(assembly, stack, target_stack_list)

        # final step to get the inputs to this instruction ordered
        # correctly on the stack
        self._stack_reorder(assembly, stack, operands)  # type: ignore

        # some instructions (i.e. invoke) need to do stack manipulations
        # with the stack model containing the return value(s), so we fiddle
        # with the stack model beforehand.

        # Step 4: Push instruction's return value to stack
        stack.pop(len(operands))
        if inst.output is not None:
            stack.push(inst.output)

        # Step 5: Emit the EVM instruction(s)
        if opcode in _ONE_TO_ONE_INSTRUCTIONS:
            assembly.append(opcode.upper())
        elif opcode == "alloca":
            pass
        elif opcode == "param":
            pass
        elif opcode == "store":
            pass
        elif opcode == "dbname":
            pass
        elif opcode in ["codecopy", "dloadbytes"]:
            assembly.append("CODECOPY")
        elif opcode == "jnz":
            # jump if not zero
            if_nonzero_label = inst.operands[1]
            if_zero_label = inst.operands[2]
            assembly.append(f"_sym_{if_nonzero_label.value}")
            assembly.append("JUMPI")

            # make sure the if_zero_label will be optimized out
            # assert if_zero_label == next(iter(inst.parent.cfg_out)).label

            assembly.append(f"_sym_{if_zero_label.value}")
            assembly.append("JUMP")

        elif opcode == "jmp":
            assert isinstance(inst.operands[0], IRLabel)
            assembly.append(f"_sym_{inst.operands[0].value}")
            assembly.append("JUMP")
        elif opcode == "djmp":
            assert isinstance(
                inst.operands[0], IRVariable
            ), f"Expected IRVariable, got {inst.operands[0]}"
            assembly.append("JUMP")
        elif opcode == "gt":
            assembly.append("GT")
        elif opcode == "lt":
            assembly.append("LT")
        elif opcode == "invoke":
            target = inst.operands[0]
            assert isinstance(
                target, IRLabel
            ), f"invoke target must be a label (is ${type(target)} ${target})"
            assembly.extend(
                [
                    f"_sym_label_ret_{self.label_counter}",
                    f"_sym_{target.value}",
                    "JUMP",
                    f"_sym_label_ret_{self.label_counter}",
                    "JUMPDEST",
                ]
            )
            self.label_counter += 1
        elif opcode == "ret":
            assembly.append("JUMP")
        elif opcode == "return":
            assembly.append("RETURN")
        elif opcode == "exit":
            assembly.extend(["_sym__ctor_exit", "JUMP"])
        elif opcode == "phi":
            pass
        elif opcode == "sha3":
            assembly.append("SHA3")
        elif opcode == "sha3_64":
            assembly.extend(
                [
                    *PUSH(MemoryPositions.FREE_VAR_SPACE),
                    "MSTORE",
                    *PUSH(MemoryPositions.FREE_VAR_SPACE2),
                    "MSTORE",
                    *PUSH(64),
                    *PUSH(MemoryPositions.FREE_VAR_SPACE),
                    "SHA3",
                ]
            )
        elif opcode == "ceil32":
            assembly.extend([*PUSH(31), "ADD", *PUSH(31), "NOT", "AND"])
        elif opcode == "assert":
            assembly.extend(["ISZERO", "_sym___revert", "JUMPI"])
        elif opcode == "assert_unreachable":
            end_symbol = mksymbol("reachable")
            assembly.extend([end_symbol, "JUMPI", "INVALID", end_symbol, "JUMPDEST"])
        elif opcode == "iload":
            addr = inst.operands[0]
            if isinstance(addr, IRLiteral):
                assembly.extend(["_OFST", "_mem_deploy_end", addr.value])
            else:
                assembly.extend(["_mem_deploy_end", "ADD"])
            assembly.append("MLOAD")
        elif opcode == "istore":
            addr = inst.operands[1]
            if isinstance(addr, IRLiteral):
                assembly.extend(["_OFST", "_mem_deploy_end", addr.value])
            else:
                assembly.extend(["_mem_deploy_end", "ADD"])
            assembly.append("MSTORE")
        elif opcode == "log":
            assembly.extend([f"LOG{log_topic_count}"])
        elif opcode == "nop":
            pass
        else:
            raise Exception(f"Unknown opcode: {opcode}")

        # Step 6: Emit instructions output operands (if any)
        if inst.output is not None:
            if "call" in inst.opcode and inst.output not in next_liveness:
                self.pop(assembly, stack)

        return apply_line_numbers(inst, assembly)

    def pop(self, assembly, stack, num=1):
        stack.pop(num)
        assembly.extend(["POP"] * num)

    def swap(self, assembly, stack, depth):
        # Swaps of the top is no op
        if depth == 0:
            return

        stack.swap(depth)
        assembly.append(_evm_swap_for(depth))

    def dup(self, assembly, stack, depth):
        stack.dup(depth)
        assembly.append(_evm_dup_for(depth))

    def swap_op(self, assembly, stack, op):
        self.swap(assembly, stack, stack.get_depth(op))

    def dup_op(self, assembly, stack, op):
        self.dup(assembly, stack, stack.get_depth(op))


def _evm_swap_for(depth: int) -> str:
    swap_idx = -depth
    if not (1 <= swap_idx <= 16):
        raise StackTooDeep(f"Unsupported swap depth {swap_idx}")
    return f"SWAP{swap_idx}"


def _evm_dup_for(depth: int) -> str:
    dup_idx = 1 - depth
    if not (1 <= dup_idx <= 16):
        raise StackTooDeep(f"Unsupported dup depth {dup_idx}")
    return f"DUP{dup_idx}"
